#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

from titus.fcn import Fcn
from titus.fcn import LibFcn
from titus.signature import Sig
from titus.signature import Sigs
from titus.datatype import *
from titus.errors import *
from titus.util import callfcn, div
import titus.P as P
from titus.lib.array import argLowestN
from titus.lib.core import powLikeJava

provides = {}
def provide(fcn):
    provides[fcn.name] = fcn

prefix = "m.kernel."

#################################################################### 
def dot(x, y, code, fcnName, pos):
    if len(x) != len(y):
        raise PFARuntimeException("arrays must have same length", code, fcnName, pos)
    else:
        return sum([i*j for i, j in zip(x,y)])

class RBF(LibFcn):
    name = prefix + "rbf"
    sig = Sig([{"x": P.Array(P.Double())}, 
               {"y": P.Array(P.Double())},
               {"gamma": P.Double()}], P.Double())
     
    errcodeBase = 23010
    def __call__(self, state, scope, pos, paramTypes, x, y, gamma):
        if len(x) != len(y):
            raise PFARuntimeException("arrays must have same length", self.errcodeBase + 0, self.name, pos)
        diff = [abs(i - j)**2 for i,j in zip(x,y)]
        return math.exp(-gamma*sum(diff))
provide(RBF())

class Linear(LibFcn):
    name = prefix + "linear"
    sig = Sig([{"x": P.Array(P.Double())}, 
               {"y": P.Array(P.Double())}], P.Double())
    errcodeBase = 23000
    def __call__(self, state, scope, pos, paramTypes, x, y):
        return dot(x, y, self.errcodeBase + 0, self.name, pos)
provide(Linear())

class Poly(LibFcn):
    name = prefix + "poly"
    sig = Sig([{"x": P.Array(P.Double())}, 
               {"y": P.Array(P.Double())},
               {"gamma": P.Double()},
               {"intercept": P.Double()},
               {"degree": P.Double()}], P.Double())
    errcodeBase = 23020
    def __call__(self, state, scope, pos, paramTypes, x, y, gamma, intercept, degree):
        return powLikeJava(gamma*dot(x, y, self.errcodeBase + 0, self.name, pos) + intercept, degree)
provide(Poly())

class Sigmoid(LibFcn):
    name = prefix + "sigmoid"
    sig = Sig([{"x": P.Array(P.Double())}, 
               {"y": P.Array(P.Double())},
               {"gamma": P.Double()},
               {"intercept": P.Double()}], P.Double())
    errcodeBase = 23030
    def __call__(self, state, scope, pos, paramTypes, x, y, gamma, intercept):
        return math.tanh(gamma * dot(x, y, self.errcodeBase + 0, self.name, pos) + intercept)
provide(Sigmoid())



#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from titus.fcn import Fcn
from titus.fcn import LibFcn
from titus.signature import Sig
from titus.signature import Sigs
from titus.datatype import *
from titus.errors import *
from titus.util import callfcn, negativeIndex, startEnd, stringToBytes, bytesToString
import titus.P as P

import sys
import ctypes

provides = {}
def provide(fcn):
    provides[fcn.name] = fcn

prefix = "re."

####################################################################

## try to import clib (give warning if user doesnt have it)
## wrap clib for either ubuntu linux or mac after checking system
#    - works on sys.platform -> darwin, sys.platform -> linux
## expose clib regex functionality to Regexer

# this class holds info about clib regex compile settings
# for use in the future to dynamically wrap clibs regex
class RegexSpecs(object):
    def __init__(self):
        system = sys.platform
        if system == "linux":
            # tested on
            self.libname = "libc.so.6"
            self._linuxSpecs()
            self.importSuccessfull = True
        elif system == "darwin":
            # tested on
            self.libname = "/usr/lib/libc.dylib"
            self._darwinSpecs()
            self.importSuccessfull = True
        else:
            self.importSuccessfull = False

    def _darwinSpecs(self):
        self.multilineFlag = 1
        self.posixExtendedSyntaxFlag = 1
        self.numNullPointersBefore_re_nsub = 1
        self.numNullPointersAfter_re_nsub = 2
        self.field_re_nsub = ("re_nsub", ctypes.c_size_t)
        self.field_rm_so = ("rm_so", ctypes.c_ulong)
        self.field_rm_co = ("rm_co", ctypes.c_ulong)

    def _linuxSpecs(self):
        self.multilineFlag = 1
        self.posixExtendedSyntaxFlag = 1
        self.numNullPointersBefore_re_nsub = 6
        self.numNullPointersAfter_re_nsub = 1
        self.field_re_nsub = ("re_nsub", ctypes.c_size_t)
        self.field_rm_so = ("rm_so", ctypes.c_int)
        self.field_rm_co = ("rm_co", ctypes.c_int)
        self.importSuccessfull = True

class Regexer(object):
    firstTime = True
    clibSpecs = None
    Regex_t = None
    Regmatch_t = None
    libc = None

    def __init__(self, haystack, pattern, code, name, pos):
        if Regexer.firstTime:
            # get clib regex specs for the wrapper
            Regexer.clibSpecs = RegexSpecs()

            # this block is here to run only when titus evaluates a pfa regex function
            if not Regexer.clibSpecs.importSuccessfull:
                raise ImportError("clib unavailable")
            else:
                # define the uninstantiated regex_t class
                regex_t_fields = []
                for i in range(0, Regexer.clibSpecs.numNullPointersBefore_re_nsub):
                    regex_t_fields.append( ("unusedname", ctypes.c_void_p) )
                regex_t_fields.append(Regexer.clibSpecs.field_re_nsub)
                for i in range(0, Regexer.clibSpecs.numNullPointersAfter_re_nsub):
                    regex_t_fields.append( ("unusedname", ctypes.c_void_p) )
                Regexer.Regex_t = type("Regex_t", (ctypes.Structure,), {"_fields_": regex_t_fields})
                # define the uninstantiated regmatch_t class
                regmatch_t_fields = [Regexer.clibSpecs.field_rm_so, Regexer.clibSpecs.field_rm_co]
                Regexer.Regmatch_t = type("Regmatch_t", (ctypes.Structure,), {"_fields_": regmatch_t_fields})
                # actually import the clibrary
                Regexer.libc = ctypes.cdll.LoadLibrary(Regexer.clibSpecs.libname)

            Regexer.firstTime = False

        # haystack and pattern come in as type(haystack) == unicode
        self.haystack = haystack
        # keep track of haystack unicode indexes
        self.haystack_indices = utf8ByteIndexes(str(haystack))
        # compile pattern into regex_t object
        regex_t = Regexer.Regex_t()
        try:
            comp = Regexer.libc.regcomp(ctypes.byref(regex_t), pattern.encode("utf-8"), Regexer.clibSpecs.posixExtendedSyntaxFlag)
        except AttributeError:
            comp = Regexer.libc.regcomp(ctypes.byref(regex_t), pattern, Regexer.clibSpecs.posixExtendedSyntaxFlag)

        if (comp != 0):
            raise PFARuntimeException("bad pattern", code, name, pos)
        self.regex_t = regex_t

        self.numGroups = int(self.regex_t.re_nsub) + 1
        self.groupArray = (Regexer.Regmatch_t * self.numGroups)()
        self.indexOffset = 0

    def search(self, start):
        ex = Regexer.libc.regexec(ctypes.byref(self.regex_t), self.haystack[start:],
                          self.numGroups, self.groupArray, 0)
        self.indexOffset = start
        if (ex != 0) or (start >= len(self.haystack)):
            return False
        else:
            return True

    def groupsFound(self):
        # count results in group array that arent (-1, -1)
        numFound = 0
        for i in range(0, self.numGroups):
            if self.groupArray[i].rm_so != -1:
                numFound += 1
        return numFound

    def getRegion(self):
        return Region(self.groupArray, self.indexOffset)

    def free(self):
        # free after every use!
        Regexer.libc.regfree(ctypes.byref(self.regex_t))

# region class (for use similar to joni in scala)
class Region(object):
    def __init__(self, groupArray, indexOffset):
        beg = []
        end = []
        for i in range(0, len(groupArray)):
            if groupArray[i].rm_so != -1:
                beg.append(int(groupArray[i].rm_so) + indexOffset)
                end.append(int(groupArray[i].rm_co) + indexOffset)
        self.beg = beg
        self.end = end

def utf8ByteIndexes(s):
    out = []
    cumulative = 0
    for i in range(len(s)):
        out.append(cumulative)
        c = ord(s[i])
        if   (c <=     0x7f): cumulative += 1
        elif (c <=    0x7ff): cumulative += 2
        elif (c <=   0xffff): cumulative += 3
        elif (c <= 0x1fffff): cumulative += 4
        else:
            raise Exception
    return out

def convert(haystack, pattern, paramType0):
    if paramType0 == "string":
        return haystack.encode("utf-8"), pattern.encode("utf-8"), lambda x: x.decode("utf-8")
    else:
        if isinstance(haystack, bytes):
            cnvtHaystack = haystack
        else:
            cnvtHaystack = stringToBytes(haystack)
        if isinstance(pattern, bytes):
            cnvtPattern = pattern
        else:
            cnvtPattern = stringToBytes(pattern)       
        return cnvtHaystack, cnvtPattern, lambda x: bytesToString(x)

############################################################# Index
class Index(LibFcn):
    name = prefix + "index"
    sig = Sigs([Sig([{"haystack": P.String()}, {"pattern": P.String()}], P.Array(P.Int())),
               Sig([{"haystack": P.Bytes()},  {"pattern": P.Bytes()}],  P.Array(P.Int()))])
    errcodeBase = 35000
    def __call__(self, state, scope, pos, paramTypes, haystack, pattern):
        haystack, pattern, to = convert(haystack, pattern, paramTypes[0])
        re = Regexer(haystack, pattern, self.errcodeBase + 0, self.name, pos)
        if re.search(0):
            region = re.getRegion()
            out = [region.beg[0], region.end[0]]
        else:
            out = []
        re.free()
        return out
provide(Index())

############################################################# Contains
class Contains(LibFcn):
    name = prefix + "contains"
    sig = Sigs([Sig([{"haystack": P.String()}, {"pattern": P.String()}], P.Boolean()),
                Sig([{"haystack": P.Bytes()},  {"pattern": P.Bytes()}],  P.Boolean())])
    errcodeBase = 35010
    def __call__(self, state, scope, pos, paramTypes, haystack, pattern):
        haystack, pattern, to = convert(haystack, pattern, paramTypes[0])
        re = Regexer(haystack, pattern, self.errcodeBase + 0, self.name, pos)
        out = re.search(0)
        re.free()
        return out
provide(Contains())

############################################################# Count
class Count(LibFcn):
    name = prefix + "count"
    sig = Sigs([Sig([{"haystack": P.String()}, {"pattern": P.String()}], P.Int()),
                Sig([{"haystack": P.Bytes()},  {"pattern": P.Bytes()}],  P.Int())])
    errcodeBase = 35020
    def __call__(self, state, scope, pos, paramTypes, haystack, pattern):
        haystack, pattern, to = convert(haystack, pattern, paramTypes[0])
        re = Regexer(haystack, pattern, self.errcodeBase + 0, self.name, pos)
        total = 0
        found = re.search(0)
        region = re.getRegion()
        start = region.end[0]
        while found:
            total += 1
            found = re.search(start)
            region = re.getRegion()
            start = region.end[0]
        re.free()
        return total
provide(Count())

############################################################# Rindex
class RIndex(LibFcn):
    name = prefix + "rindex"
    sig = Sigs([Sig([{"haystack": P.String()}, {"pattern": P.String()}], P.Array(P.Int())),
                Sig([{"haystack": P.Bytes()}, {"pattern": P.Bytes()}], P.Array(P.Int()))])
    errcodeBase = 35030
    def __call__(self, state, scope, pos, paramTypes, haystack, pattern):
        haystack, pattern, to = convert(haystack, pattern, paramTypes[0])
        re = Regexer(haystack, pattern, self.errcodeBase + 0, self.name, pos)
        found = re.search(0)
        region = re.getRegion()
        start = 0
        if found:
            while found:
                region = re.getRegion()
                start = region.end[0]
                found = re.search(start)
            out = [region.beg[0], region.end[0]]
        else:
            out = []
        re.free()
        return out
provide(RIndex())

############################################################# Groups
class Groups(LibFcn):
    name = prefix + "groups"
    sig = Sigs([Sig([{"haystack": P.String()}, {"pattern": P.String()}], P.Array(P.Array(P.Int()))),
                Sig([{"haystack": P.Bytes()}, {"pattern": P.Bytes()}], P.Array(P.Array(P.Int())))])
    errcodeBase = 35040
    def __call__(self, state, scope, pos, paramTypes, haystack, pattern):
        haystack, pattern, to = convert(haystack, pattern, paramTypes[0])
        re = Regexer(haystack, pattern, self.errcodeBase + 0, self.name, pos)
        found = re.search(0)
        region = re.getRegion()
        start = region.end[0]
        out = []
        if found:
            for i in range(0, re.groupsFound()):
                out.append([region.beg[i], region.end[i]])
        else:
            out = []
        re.free()
        return out
provide(Groups())

############################################################# IndexAll
class IndexAll(LibFcn):
    name = prefix + "indexall"
    sig = Sigs([Sig([{"haystack": P.String()}, {"pattern": P.String()}], P.Array(P.Array(P.Int()))),
                Sig([{"haystack": P.Bytes()}, {"pattern": P.Bytes()}], P.Array(P.Array(P.Int())))])
    errcodeBase = 35050
    def __call__(self, state, scope, pos, paramTypes, haystack, pattern):
        haystack, pattern, to = convert(haystack, pattern, paramTypes[0])
        re = Regexer(haystack, pattern, self.errcodeBase + 0, self.name, pos)
        found = re.search(0)
        region = re.getRegion()
        start = region.end[0]
        out = []
        if found:
            while found:
                region = re.getRegion()
                start = region.end[0]
                found = re.search(start)
                out.append([region.beg[0], region.end[0]])
        else:
            out = []
        re.free()
        return out
provide(IndexAll())

############################################################# FindAll
class FindAll(LibFcn):
    name = prefix + "findall"
    sig = Sigs([Sig([{"haystack": P.String()}, {"pattern": P.String()}], P.Array(P.String())),
                Sig([{"haystack": P.Bytes()}, {"pattern": P.Bytes()}], P.Array(P.Bytes()))])
    errcodeBase = 35060
    def __call__(self, state, scope, pos, paramTypes, haystack, pattern):
        haystack, pattern, to = convert(haystack, pattern, paramTypes[0])
        re = Regexer(haystack, pattern, self.errcodeBase + 0, self.name, pos)
        found = re.search(0)
        region = re.getRegion()
        start = region.end[0]
        out = []
        if found:
            while found:
                region = re.getRegion()
                start = region.end[0]
                out.append(to(haystack[region.beg[0]:region.end[0]]))
                found = re.search(start)
        else:
            out = []
        re.free()
        return out
provide(FindAll())

############################################################# FindFirst
class FindFirst(LibFcn):
    name = prefix + "findfirst"
    sig = Sigs([Sig([{"haystack": P.String()}, {"pattern": P.String()}], P.Union([P.String(), P.Null()])),
                Sig([{"haystack": P.Bytes()}, {"pattern": P.Bytes()}], P.Union([P.Bytes(), P.Null()]))])
    errcodeBase = 35070
    def __call__(self, state, scope, pos, paramTypes, haystack, pattern):
        haystack, pattern, to = convert(haystack, pattern, paramTypes[0])
        re = Regexer(haystack, pattern, self.errcodeBase + 0, self.name, pos)
        found = re.search(0)
        if found:
            region = re.getRegion()
            out = to(haystack[region.beg[0]:region.end[0]])
        else:
            out = None
        re.free()
        if out is not None:
            return {paramTypes[0]: out}
        else:
            return out
provide(FindFirst())

############################################################# FindGroupsFirst
class FindGroupsFirst(LibFcn):
    name = prefix + "findgroupsfirst"
    sig = Sigs([Sig([{"haystack": P.String()}, {"pattern": P.String()}], P.Array(P.String())),
                Sig([{"haystack": P.Bytes()}, {"pattern": P.Bytes()}], P.Array(P.Bytes()))])
    errcodeBase = 35080
    def __call__(self, state, scope, pos, paramTypes, haystack, pattern):
        haystack, pattern, to = convert(haystack, pattern, paramTypes[0])
        re = Regexer(haystack, pattern, self.errcodeBase + 0, self.name, pos)
        start = 0
        found = re.search(start)
        out = []
        region = re.getRegion()
        if (found):
            for i in range(0,re.groupsFound()):
                out.append(to(haystack[region.beg[i]:region.end[i]]))
        else:
            out = []
        re.free()
        return out
provide(FindGroupsFirst())

############################################################# FindGroupsAll
class FindGroupsAll(LibFcn):
    name = prefix + "findgroupsall"
    sig = Sigs([Sig([{"haystack": P.String()}, {"pattern": P.String()}], P.Array(P.Array(P.String()))),
                Sig([{"haystack": P.Bytes()}, {"pattern": P.Bytes()}], P.Array(P.Array(P.Bytes())))])
    errcodeBase = 35090
    def __call__(self, state, scope, pos, paramTypes, haystack, pattern):
        haystack, pattern, to = convert(haystack, pattern, paramTypes[0])
        re = Regexer(haystack, pattern, self.errcodeBase + 0, self.name, pos)
        start = 0
        found = re.search(start)
        region = re.getRegion()
        out = []
        if found:
            while found:
                region = re.getRegion()
                groupList = []
                for i in range(0, re.groupsFound()):
                    groupList.append(to(haystack[region.beg[i]:region.end[i]]))
                out.append(groupList)
                start = region.end[0]
                found = re.search(start)
        else:
            out = []
        re.free()
        return out
provide(FindGroupsAll())

############################################################# GroupsAll
class GroupsAll(LibFcn):
    name = prefix + "groupsall"
    sig = Sigs([Sig([{"haystack": P.String()}, {"pattern": P.String()}], P.Array(P.Array(P.Array(P.Int())))),
                Sig([{"haystack": P.Bytes()}, {"pattern": P.Bytes()}], P.Array(P.Array(P.Array(P.Int()))))])
    errcodeBase = 35100
    def __call__(self, state, scope, pos, paramTypes, haystack, pattern):
        haystack, pattern, to = convert(haystack, pattern, paramTypes[0])
        re = Regexer(haystack, pattern, self.errcodeBase + 0, self.name, pos)
        start = 0
        found = re.search(start)
        region = re.getRegion()
        out = []
        if found:
            while found:
                region = re.getRegion()
                groupList = []
                for i in range(0, re.groupsFound()):
                    groupList.append([region.beg[i], region.end[i]])
                out.append(groupList)
                start = region.end[0]
                found = re.search(start)
        else:
            out = []
        re.free()
        return out
provide(GroupsAll())

############################################################# ReplaceFirst
class ReplaceFirst(LibFcn):
    name = prefix + "replacefirst"
    sig = Sigs([Sig([{"haystack": P.String()}, {"pattern": P.String()}, {"replacement": P.String()}], P.String()),
                Sig([{"haystack": P.Bytes()}, {"pattern": P.Bytes()}, {"replacement": P.Bytes()}], P.Bytes())])
    errcodeBase = 35110
    def __call__(self, state, scope, pos, paramTypes, haystack, pattern, replacement):
        haystack, pattern, to = convert(haystack, pattern, paramTypes[0])
        if isinstance(replacement, bytes):
            replacement = bytesToString(replacement)      
        re = Regexer(haystack, pattern, self.errcodeBase + 0, self.name, pos)
        found = re.search(0)
        region = re.getRegion()
        if found:
            out = to(haystack[:region.beg[0]]) + replacement + to(haystack[region.end[0]:])
        else:
            out = to(haystack)
        re.free()
        return out
provide(ReplaceFirst())

############################################################# ReplaceLast
class ReplaceLast(LibFcn):
    name = prefix + "replacelast"
    sig = Sigs([Sig([{"haystack": P.String()}, {"pattern": P.String()}, {"replacement": P.String()}], P.String()),
                Sig([{"haystack": P.Bytes()}, {"pattern": P.Bytes()}, {"replacement": P.Bytes()}], P.Bytes())])
    errcodeBase = 35120
    def __call__(self, state, scope, pos, paramTypes, haystack, pattern, replacement):
        haystack, pattern, to = convert(haystack, pattern, paramTypes[0])
        if isinstance(replacement, bytes):
            replacement = bytesToString(replacement)
        re = Regexer(haystack, pattern, self.errcodeBase + 0, self.name, pos)
        found = re.search(0)
        region = re.getRegion()
        start = 0
        if found:
            while found:
                region = re.getRegion()
                start = region.end[0]
                found = re.search(start)
            out = to(haystack[:region.beg[0]]) + replacement + to(haystack[region.end[0]:])
        else:
            out = to(haystack)
        re.free()
        return out
provide(ReplaceLast())

############################################################# Split
class Split(LibFcn):
    name = prefix + "split"
    sig = Sigs([Sig([{"haystack": P.String()}, {"pattern": P.String()}], P.Array(P.String())),
                Sig([{"haystack": P.Bytes()}, {"pattern": P.Bytes()}], P.Array(P.Bytes()))])
    errcodeBase = 35130
    def __call__(self, state, scope, pos, paramTypes, haystack, pattern):
        haystack, pattern, to = convert(haystack, pattern, paramTypes[0])
        re = Regexer(haystack, pattern, self.errcodeBase + 0, self.name, pos)
        out = []
        start = 0
        found = re.search(start)
        region = re.getRegion()
        beg = 0
        end = region.beg[0]
        if (end == 0):
            flag = True
        else:
            flag = False
        if found:
            while found:
                out.append(to(haystack[beg:end]))
                beg = region.end[0]
                found = re.search(beg)
                region = re.getRegion()
                end = region.beg[0]
            if beg != len(haystack):
                out.append(to(haystack[beg:]))
            if flag:
                out = out[1:]
        else:
            out = [to(haystack)]
        re.free()
        return out
provide(Split())

############################################################# ReplaceAll
class ReplaceAll(LibFcn):
    name = prefix + "replaceall"
    sig = Sigs([Sig([{"haystack": P.String()}, {"pattern": P.String()}, {"replacement": P.String()}], P.String()),
                Sig([{"haystack": P.Bytes()}, {"pattern": P.Bytes()}, {"replacement": P.Bytes()}], P.Bytes())])
    errcodeBase = 35140
    def __call__(self, state, scope, pos, paramTypes, haystack, pattern, replacement):
        original = haystack
        if isinstance(original, bytes):
            original = bytesToString(original)
        haystack, pattern, to = convert(haystack, pattern, paramTypes[0])
        if isinstance(replacement, bytes):
            replacement = bytesToString(replacement)
        re = Regexer(haystack, pattern, self.errcodeBase + 0, self.name, pos)
        found = re.search(0)
        region = re.getRegion()
        beg = 0
        end = region.beg[0]
        out = ""
        if found:
            while found:
                out = out + to(haystack[beg:end]) + replacement
                beg = region.end[0]
                found = re.search(beg)
                region = re.getRegion()
                end = region.beg[0]
            if beg != len(haystack):
                out = out + to(haystack[beg:])
        else:
            out = original
        re.free()
        return out
provide(ReplaceAll())



#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

from titus.fcn import Fcn
from titus.fcn import LibFcn
from titus.signature import Sig
from titus.signature import Sigs
from titus.datatype import *
from titus.errors import *
import titus.P as P

provides = {}
def provide(fcn):
    provides[fcn.name] = fcn

prefix = "impute."

class ErrorOnNull(LibFcn):
    name = prefix + "errorOnNull"
    sig = Sig([{"x": P.Union([P.Wildcard("A"), P.Null()])}], P.Wildcard("A"))
    errcodeBase = 21000
    def __call__(self, state, scope, pos, paramTypes, x):
        if x is None:
            raise PFARuntimeException("encountered null", self.errcodeBase + 0, self.name, pos)
        else:
            if isinstance(paramTypes[-1], (list, tuple)):
                return x
            else:
                if isinstance(x, dict):
                    return list(x.values())[0]
                else:
                    return x
provide(ErrorOnNull())

class DefaultOnNull(LibFcn):
    name = prefix + "defaultOnNull"
    sig = Sig([{"x": P.Union([P.Wildcard("A"), P.Null()])}, {"default": P.Wildcard("A")}], P.Wildcard("A"))
    errcodeBase = 21010
    def __call__(self, state, scope, pos, paramTypes, x, default):
        if x is None:
            return default
        else:
            if isinstance(paramTypes[-1], (list, tuple)):
                return x
            else:
                if isinstance(x, dict):
                    return list(x.values())[0]
                else:
                    return x
provide(DefaultOnNull())

class IsNan(LibFcn):
    name = prefix + "isnan"
    sig = Sigs([Sig([{"x": P.Float()}], P.Boolean()),
                Sig([{"x": P.Double()}], P.Boolean())])
    errcodeBase = 21020
    def __call__(self, state, scope, pos, paramTypes, x):
        return math.isnan(x)
provide(IsNan())

class IsInf(LibFcn):
    name = prefix + "isinf"
    sig = Sigs([Sig([{"x": P.Float()}], P.Boolean()),
                Sig([{"x": P.Double()}], P.Boolean())])
    errcodeBase = 21030
    def __call__(self, state, scope, pos, paramTypes, x):
        return math.isinf(x)
provide(IsInf())

class IsNum(LibFcn):
    name = prefix + "isnum"
    sig = Sigs([Sig([{"x": P.Float()}], P.Boolean()),
                Sig([{"x": P.Double()}], P.Boolean())])
    errcodeBase = 21040
    def __call__(self, state, scope, pos, paramTypes, x):
        return not math.isnan(x) and not math.isinf(x)
provide(IsNum())

class ErrorOnNonNum(LibFcn):
    name = prefix + "errorOnNonNum"
    sig = Sigs([Sig([{"x": P.Float()}], P.Float()),
                Sig([{"x": P.Double()}], P.Double())])
    errcodeBase = 21050
    def __call__(self, state, scope, pos, paramTypes, x):
        if math.isnan(x):
            raise PFARuntimeException("encountered nan", self.errcodeBase + 0, self.name, pos)
        elif math.isinf(x):
            if x > 0.0:
                raise PFARuntimeException("encountered +inf", self.errcodeBase + 1, self.name, pos)
            else:
                raise PFARuntimeException("encountered -inf", self.errcodeBase + 2, self.name, pos)
        else:
            return x
provide(ErrorOnNonNum())

class DefaultOnNonNum(LibFcn):
    name = prefix + "defaultOnNonNum"
    sig = Sigs([Sig([{"x": P.Float()}, {"default": P.Float()}], P.Float()),
                Sig([{"x": P.Double()}, {"default": P.Double()}], P.Double())])
    errcodeBase = 21060
    def __call__(self, state, scope, pos, paramTypes, x, default):
        if math.isnan(x) or math.isinf(x):
            return default
        else:
            return x
provide(DefaultOnNonNum())



#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import codecs
import re

from titus.fcn import Fcn
from titus.fcn import LibFcn
from titus.signature import Sig
from titus.signature import Sigs
from titus.datatype import *
from titus.errors import *
from titus.util import startEnd, stringToBytes, bytesToString
import titus.P as P

provides = {}
def provide(fcn):
    provides[fcn.name] = fcn

prefix = "bytes."

#################################################################### basic access

class Len(LibFcn):
    name = prefix + "len"
    sig = Sig([{"x": P.Bytes()}], P.Int())
    errcodeBase = 16000
    def genpy(self, paramTypes, args, pos):
        return "len({0})".format(*args)
    def __call__(self, state, scope, pos, paramTypes, x):
        return len(x)
provide(Len())

class Subseq(LibFcn):
    name = prefix + "subseq"
    sig = Sig([{"x": P.Bytes()}, {"start": P.Int()}, {"end": P.Int()}], P.Bytes())
    errcodeBase = 16010
    def __call__(self, state, scope, pos, paramTypes, x, start, end):
        if isinstance(x, str):
            return x[start:end]
        else:
            return bytesToString(x[start:end])
provide(Subseq())

class SubseqTo(LibFcn):
    name = prefix + "subseqto"
    sig = Sig([{"x": P.Bytes()}, {"start": P.Int()}, {"end": P.Int()}, {"replacement": P.Bytes()}], P.Bytes())
    errcodeBase = 16020
    def __call__(self, state, scope, pos, paramTypes, x, start, end, replacement):
        if isinstance(replacement, str):
            replacement = stringToBytes(replacement)
        if isinstance(x, str):
            x = stringToBytes(x)
        normStart, normEnd = startEnd(len(x), start, end)
        before = x[:normStart]
        after = x[normEnd:]
        return bytesToString(before + replacement + after)
provide(SubseqTo())

#################################################################### testers

class Tester(LibFcn):
    sig = Sigs([Sig([{"x": P.Bytes()}], P.Boolean()),
                Sig([{"x": P.String()}], P.Boolean())])
    codec = None
    def __call__(self, state, scope, pos, paramTypes, x):
        if paramTypes[0] == "bytes" or paramTypes[0] == {"type": "bytes"}:
            try:
                codecs.decode(x, self.codec, "strict")
            except UnicodeDecodeError:
                return False
            else:
                return True
        else:
            try:
                codecs.encode(x, self.codec, "strict")
            except UnicodeEncodeError:
                return False
            else:
                return True

class IsAscii(Tester):
    name = prefix + "isAscii"
    codec = "ascii"
    errcodeBase = 16030
provide(IsAscii())

class IsLatin1(Tester):
    name = prefix + "isLatin1"
    codec = "latin_1"
    errcodeBase = 16040
provide(IsLatin1())

class isUtf8(Tester):
    name = prefix + "isUtf8"
    codec = "utf_8"
    errcodeBase = 16050
provide(isUtf8())

class isUtf16(Tester):
    name = prefix + "isUtf16"
    codec = "utf_16"
    errcodeBase = 16060
provide(isUtf16())

class isUtf16BE(Tester):
    name = prefix + "isUtf16be"
    codec = "utf_16_be"
    errcodeBase = 16070
provide(isUtf16BE())

class isUtf16LE(Tester):
    name = prefix + "isUtf16le"
    codec = "utf_16_le"
    errcodeBase = 16080
provide(isUtf16LE())

#################################################################### decoders

class Decoder(LibFcn):
    sig = Sig([{"x": P.Bytes()}], P.String())
    codec = None
    def __call__(self, state, scope, pos, paramTypes, x):
        try:
            if isinstance(x, str):
                x = stringToBytes(x)
            return codecs.decode(x, self.codec, "strict")
        except UnicodeDecodeError as err:
            raise PFARuntimeException("invalid bytes", self.errcodeBase + 0, self.name, pos)

class DecodeAscii(Decoder):
    name = prefix + "decodeAscii"
    codec = "ascii"
    errcodeBase = 16090
provide(DecodeAscii())

class DecodeLatin1(Decoder):
    name = prefix + "decodeLatin1"
    codec = "latin_1"
    errcodeBase = 16100
provide(DecodeLatin1())

class decodeUtf8(Decoder):
    name = prefix + "decodeUtf8"
    codec = "utf_8"
    errcodeBase = 16110
provide(decodeUtf8())

class decodeUtf16(Decoder):
    name = prefix + "decodeUtf16"
    codec = "utf_16"
    errcodeBase = 16120
provide(decodeUtf16())

class decodeUtf16BE(Decoder):
    name = prefix + "decodeUtf16be"
    codec = "utf_16_be"
    errcodeBase = 16130
provide(decodeUtf16BE())

class decodeUtf16LE(Decoder):
    name = prefix + "decodeUtf16le"
    codec = "utf_16_le"
    errcodeBase = 16140
provide(decodeUtf16LE())

#################################################################### encoders

class Encoder(LibFcn):
    sig = Sig([{"x": P.String()}], P.Bytes())
    codec = None
    def __call__(self, state, scope, pos, paramTypes, x):
        try:
            out_bytes = codecs.encode(x, self.codec, "strict")
            return bytesToString(out_bytes)
        except UnicodeEncodeError:
            raise PFARuntimeException("invalid string", self.errcodeBase + 0, self.name, pos)

class EncodeAscii(Encoder):
    name = prefix + "encodeAscii"
    codec = "ascii"
    errcodeBase = 16150
provide(EncodeAscii())

class EncodeLatin1(Encoder):
    name = prefix + "encodeLatin1"
    codec = "latin_1"
    errcodeBase = 16160
provide(EncodeLatin1())

class EncodeUtf8(Encoder):
    name = prefix + "encodeUtf8"
    codec = "utf_8"
    errcodeBase = 16170
provide(EncodeUtf8())

class EncodeUtf16(Encoder):
    name = prefix + "encodeUtf16"
    codec = "utf_16"
    errcodeBase = 16180
provide(EncodeUtf16())

class EncodeUtf16BE(Encoder):
    name = prefix + "encodeUtf16be"
    codec = "utf_16_be"
    errcodeBase = 16190
provide(EncodeUtf16BE())

class EncodeUtf16LE(Encoder):
    name = prefix + "encodeUtf16le"
    codec = "utf_16_le"
    errcodeBase = 16200
provide(EncodeUtf16LE())

#################################################################### base64

class ToBase64(LibFcn):
    name = prefix + "toBase64"
    sig = Sig([{"x": P.Bytes()}], P.String())
    errcodeBase = 16210
    def __call__(self, state, scope, pos, paramTypes, x):
        if isinstance(x, str):
            x = stringToBytes(x)
        return base64.b64encode(x).decode()
provide(ToBase64())

class FromBase64(LibFcn):
    name = prefix + "fromBase64"
    sig = Sig([{"s": P.String()}], P.Bytes())
    errcodeBase = 16220
    def __call__(self, state, scope, pos, paramTypes, s):
        try:
            if re.match("^[A-Za-z0-9\+/]*=*$", s) is None:
                raise TypeError
            out = base64.b64decode(s)
            return bytesToString(out)
        except TypeError:
            raise PFARuntimeException("invalid base64", self.errcodeBase + 0, self.name, pos)
provide(FromBase64())



#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import unittest

from titus.genpy import PFAEngine

class TestDumpstate(unittest.TestCase):
    def testPrivateCellsInt(self):
        engine, = PFAEngine.fromYaml('''
input: int
output: "null"
cells:
  test:
    type: int
    init: 0
action:
  - cell: test
    to:
      params: [{x: int}]
      ret: int
      do: {+: [x, input]}
  - null
''')
        self.assertEqual(engine.snapshot().cells["test"].init, "0")
        engine.action(3)
        self.assertEqual(engine.snapshot().cells["test"].init, "3")
        engine.action(2)
        self.assertEqual(engine.snapshot().cells["test"].init, "5")

    def testPrivateCellsString(self):
        engine, = PFAEngine.fromYaml('''
input: string
output: "null"
cells:
  test:
    type: string
    init: ""
action:
  - cell: test
    to:
      params: [{x: string}]
      ret: string
      do: {s.concat: [x, input]}
  - null
''')
        self.assertEqual(engine.snapshot().cells["test"].init, '""')
        engine.action("hey")
        self.assertEqual(engine.snapshot().cells["test"].init, '"hey"')
        engine.action("there")
        self.assertEqual(engine.snapshot().cells["test"].init, '"heythere"')

    def testPrivateCellsArray(self):
        engine, = PFAEngine.fromYaml('''
input: int
output: "null"
cells:
  test:
    type: {type: array, items: int}
    init: []
action:
  - cell: test
    to:
      params: [{x: {type: array, items: int}}]
      ret: {type: array, items: int}
      do: {a.append: [x, input]}
  - null
''')
        self.assertEqual(engine.snapshot().cells["test"].init, "[]")
        engine.action(3)
        self.assertEqual(engine.snapshot().cells["test"].init, "[3]")
        engine.action(2)
        self.assertEqual(engine.snapshot().cells["test"].init, "[3, 2]")

    def testPrivateCellsMap(self):
        engine, = PFAEngine.fromYaml('''
input: int
output: "null"
cells:
  test:
    type: {type: map, values: int}
    init: {"a": 0, "b": 0}
action:
  - cell: test
    path: [{string: b}]
    to:
      params: [{x: int}]
      ret: int
      do: {+: [x, input]}
  - null
''')
        self.assertEqual(json.loads(engine.snapshot().cells["test"].init)["b"], 0)
        engine.action(3)
        self.assertEqual(json.loads(engine.snapshot().cells["test"].init)["b"], 3)
        engine.action(2)
        self.assertEqual(json.loads(engine.snapshot().cells["test"].init)["b"], 5)

    def testPrivateCellsRecord(self):
        engine, = PFAEngine.fromYaml('''
input: int
output: "null"
cells:
  test:
    type:
      type: record
      name: MyRecord
      fields:
        - {name: a, type: int}
        - {name: b, type: string}
    init: {a: 0, b: hey}
action:
  - cell: test
    path: [{string: a}]
    to:
      params: [{x: int}]
      ret: int
      do: {+: [x, input]}
  - null
''')
        self.assertEqual(json.loads(engine.snapshot().cells["test"].init)["a"], 0)
        engine.action(3)
        self.assertEqual(json.loads(engine.snapshot().cells["test"].init)["a"], 3)
        engine.action(2)
        self.assertEqual(json.loads(engine.snapshot().cells["test"].init)["a"], 5)

    def testPrivateCellsUnion(self):
        engine, = PFAEngine.fromYaml('''
input: int
output: "null"
cells:
  test:
    type: [int, string]
    init: {int: 0}
action:
  - cell: test
    to:
      params: [{x: [int, string]}]
      ret: [int, string]
      do:
        cast: x
        cases:
          - as: int
            named: y
            do: {+: [y, input]}
          - as: string
            named: y
            do: y
  - null
''')
        self.assertEqual(engine.snapshot().cells["test"].init, '{"int": 0}')
        engine.action(3)
        self.assertEqual(engine.snapshot().cells["test"].init, '3')
        engine.action(2)
        self.assertEqual(engine.snapshot().cells["test"].init, '5')

    def testPublicCellsInt(self):
        engine, = PFAEngine.fromYaml('''
input: int
output: "null"
cells:
  test:
    type: int
    init: 0
    shared: true
action:
  - cell: test
    to:
      params: [{x: int}]
      ret: int
      do: {+: [x, input]}
  - null
''')
        self.assertEqual(engine.snapshot().cells["test"].init, "0")
        engine.action(3)
        self.assertEqual(engine.snapshot().cells["test"].init, "3")
        engine.action(2)
        self.assertEqual(engine.snapshot().cells["test"].init, "5")

    def testPublicCellsString(self):
        engine, = PFAEngine.fromYaml('''
input: string
output: "null"
cells:
  test:
    type: string
    init: ""
    shared: true
action:
  - cell: test
    to:
      params: [{x: string}]
      ret: string
      do: {s.concat: [x, input]}
  - null
''')
        self.assertEqual(engine.snapshot().cells["test"].init, '""')
        engine.action("hey")
        self.assertEqual(engine.snapshot().cells["test"].init, '"hey"')
        engine.action("there")
        self.assertEqual(engine.snapshot().cells["test"].init, '"heythere"')

    def testPublicCellsArray(self):
        engine, = PFAEngine.fromYaml('''
input: int
output: "null"
cells:
  test:
    type: {type: array, items: int}
    init: []
    shared: true
action:
  - cell: test
    to:
      params: [{x: {type: array, items: int}}]
      ret: {type: array, items: int}
      do: {a.append: [x, input]}
  - null
''')
        self.assertEqual(engine.snapshot().cells["test"].init, "[]")
        engine.action(3)
        self.assertEqual(engine.snapshot().cells["test"].init, "[3]")
        engine.action(2)
        self.assertEqual(engine.snapshot().cells["test"].init, "[3, 2]")

    def testPublicCellsMap(self):
        engine, = PFAEngine.fromYaml('''
input: int
output: "null"
cells:
  test:
    type: {type: map, values: int}
    init: {"a": 0, "b": 0}
    shared: true
action:
  - cell: test
    path: [{string: b}]
    to:
      params: [{x: int}]
      ret: int
      do: {+: [x, input]}
  - null
''')
        self.assertEqual(json.loads(engine.snapshot().cells["test"].init)["b"], 0)
        engine.action(3)
        self.assertEqual(json.loads(engine.snapshot().cells["test"].init)["b"], 3)
        engine.action(2)
        self.assertEqual(json.loads(engine.snapshot().cells["test"].init)["b"], 5)

    def testPublicCellsRecord(self):
        engine, = PFAEngine.fromYaml('''
input: int
output: "null"
cells:
  test:
    type:
      type: record
      name: MyRecord
      fields:
        - {name: a, type: int}
        - {name: b, type: string}
    init: {a: 0, b: hey}
    shared: true
action:
  - cell: test
    path: [{string: a}]
    to:
      params: [{x: int}]
      ret: int
      do: {+: [x, input]}
  - null
''')
        self.assertEqual(json.loads(engine.snapshot().cells["test"].init)["a"], 0)
        engine.action(3)
        self.assertEqual(json.loads(engine.snapshot().cells["test"].init)["a"], 3)
        engine.action(2)
        self.assertEqual(json.loads(engine.snapshot().cells["test"].init)["a"], 5)

    def testPublicCellsUnion(self):
        engine, = PFAEngine.fromYaml('''
input: int
output: "null"
cells:
  test:
    type: [int, string]
    init: {int: 0}
    shared: true
action:
  - cell: test
    to:
      params: [{x: [int, string]}]
      ret: [int, string]
      do:
        cast: x
        cases:
          - as: int
            named: y
            do: {+: [y, input]}
          - as: string
            named: y
            do: y
  - null
''')
        self.assertEqual(engine.snapshot().cells["test"].init, '{"int": 0}')
        engine.action(3)
        self.assertEqual(engine.snapshot().cells["test"].init, '3')
        engine.action(2)
        self.assertEqual(engine.snapshot().cells["test"].init, '5')

    def testPrivatePoolsInt(self):
        engine, = PFAEngine.fromYaml('''
input: int
output: "null"
pools:
  test:
    type: int
action:
  - pool: test
    path: [{string: zzz}]
    to:
      params: [{x: int}]
      ret: int
      do: {+: [x, input]}
    init: 0
  - null
''')
        self.assertEqual(engine.snapshot().pools["test"].init, {})
        engine.action(3)
        self.assertEqual(engine.snapshot().pools["test"].init["zzz"], "3")
        engine.action(2)
        self.assertEqual(engine.snapshot().pools["test"].init["zzz"], "5")

    def testPrivatePoolsString(self):
        engine, = PFAEngine.fromYaml('''
input: string
output: "null"
pools:
  test:
    type: string
action:
  - pool: test
    path: [{string: zzz}]
    to:
      params: [{x: string}]
      ret: string
      do: {s.concat: [x, input]}
    init: {string: ""}
  - null
''')
        self.assertEqual(engine.snapshot().pools["test"].init, {})
        engine.action("hey")
        self.assertEqual(engine.snapshot().pools["test"].init["zzz"], '"hey"')
        engine.action("there")
        self.assertEqual(engine.snapshot().pools["test"].init["zzz"], '"heythere"')

    def testPrivatePoolsArray(self):
        engine, = PFAEngine.fromYaml('''
input: int
output: "null"
pools:
  test:
    type: {type: array, items: int}
action:
  - pool: test
    path: [{string: zzz}]
    to:
      params: [{x: {type: array, items: int}}]
      ret: {type: array, items: int}
      do: {a.append: [x, input]}
    init: {value: [], type: {type: array, items: int}}
  - null
''')
        self.assertEqual(engine.snapshot().pools["test"].init, {})
        engine.action(3)
        self.assertEqual(engine.snapshot().pools["test"].init["zzz"], "[3]")
        engine.action(2)
        self.assertEqual(engine.snapshot().pools["test"].init["zzz"], "[3, 2]")

    def testPrivatePoolsMap(self):
        engine, = PFAEngine.fromYaml('''
input: int
output: "null"
pools:
  test:
    type: {type: map, values: int}
    init: {"zzz": {"a": 0, "b": 0}}
action:
  - pool: test
    path: [{string: zzz}, {string: b}]
    to:
      params: [{x: int}]
      ret: int
      do: {+: [x, input]}
    init: {value: {}, type: {type: map, values: int}}
  - null
''')
        self.assertEqual(json.loads(engine.snapshot().pools["test"].init["zzz"])["b"], 0)
        engine.action(3)
        self.assertEqual(json.loads(engine.snapshot().pools["test"].init["zzz"])["b"], 3)
        engine.action(2)
        self.assertEqual(json.loads(engine.snapshot().pools["test"].init["zzz"])["b"], 5)

    def testPrivatePoolsRecord(self):
        engine, = PFAEngine.fromYaml('''
input: int
output: "null"
pools:
  test:
    type:
      type: record
      name: MyRecord
      fields:
        - {name: a, type: int}
        - {name: b, type: string}
action:
  - pool: test
    path: [{string: zzz}, {string: a}]
    to:
      params: [{x: int}]
      ret: int
      do: {+: [x, input]}
    init: {value: {a: 0, b: hey}, type: MyRecord}
  - null
''')
        self.assertEqual(engine.snapshot().pools["test"].init, {})
        engine.action(3)
        self.assertEqual(json.loads(engine.snapshot().pools["test"].init["zzz"])["a"], 3)
        engine.action(2)
        self.assertEqual(json.loads(engine.snapshot().pools["test"].init["zzz"])["a"], 5)

    def testPrivatePoolsUnion(self):
        engine, = PFAEngine.fromYaml('''
input: int
output: "null"
pools:
  test:
    type: [int, string]
action:
  - pool: test
    path: [{string: zzz}]
    to:
      params: [{x: [int, string]}]
      ret: [int, string]
      do:
        cast: x
        cases:
          - as: int
            named: y
            do: {+: [y, input]}
          - as: string
            named: y
            do: y
    init: {int: 0}
  - null
''')
        self.assertEqual(engine.snapshot().pools["test"].init, {})
        engine.action(3)
        self.assertEqual(engine.snapshot().pools["test"].init["zzz"], "3")
        engine.action(2)
        self.assertEqual(engine.snapshot().pools["test"].init["zzz"], "5")

    def testPublicPoolsInt(self):
        engine, = PFAEngine.fromYaml('''
input: int
output: "null"
pools:
  test:
    type: int
    shared: true
action:
  - pool: test
    path: [{string: zzz}]
    to:
      params: [{x: int}]
      ret: int
      do: {+: [x, input]}
    init: 0
  - null
''')
        self.assertEqual(engine.snapshot().pools["test"].init, {})
        engine.action(3)
        self.assertEqual(engine.snapshot().pools["test"].init["zzz"], "3")
        engine.action(2)
        self.assertEqual(engine.snapshot().pools["test"].init["zzz"], "5")

    def testPublicPoolsString(self):
        engine, = PFAEngine.fromYaml('''
input: string
output: "null"
pools:
  test:
    type: string
    shared: true
action:
  - pool: test
    path: [{string: zzz}]
    to:
      params: [{x: string}]
      ret: string
      do: {s.concat: [x, input]}
    init: {string: ""}
  - null
''')
        self.assertEqual(engine.snapshot().pools["test"].init, {})
        engine.action("hey")
        self.assertEqual(engine.snapshot().pools["test"].init["zzz"], '"hey"')
        engine.action("there")
        self.assertEqual(engine.snapshot().pools["test"].init["zzz"], '"heythere"')

    def testPublicPoolsArray(self):
        engine, = PFAEngine.fromYaml('''
input: int
output: "null"
pools:
  test:
    type: {type: array, items: int}
    shared: true
action:
  - pool: test
    path: [{string: zzz}]
    to:
      params: [{x: {type: array, items: int}}]
      ret: {type: array, items: int}
      do: {a.append: [x, input]}
    init: {value: [], type: {type: array, items: int}}
  - null
''')
        self.assertEqual(engine.snapshot().pools["test"].init, {})
        engine.action(3)
        self.assertEqual(engine.snapshot().pools["test"].init["zzz"], "[3]")
        engine.action(2)
        self.assertEqual(engine.snapshot().pools["test"].init["zzz"], "[3, 2]")

    def testPublicPoolsMap(self):
        engine, = PFAEngine.fromYaml('''
input: int
output: "null"
pools:
  test:
    type: {type: map, values: int}
    init: {"zzz": {"a": 0, "b": 0}}
    shared: true
action:
  - pool: test
    path: [{string: zzz}, {string: b}]
    to:
      params: [{x: int}]
      ret: int
      do: {+: [x, input]}
    init: {value: {}, type: {type: map, values: int}}
  - null
''')
        self.assertEqual(json.loads(engine.snapshot().pools["test"].init["zzz"])["b"], 0)
        engine.action(3)
        self.assertEqual(json.loads(engine.snapshot().pools["test"].init["zzz"])["b"], 3)
        engine.action(2)
        self.assertEqual(json.loads(engine.snapshot().pools["test"].init["zzz"])["b"], 5)

    def testPublicPoolsRecord(self):
        engine, = PFAEngine.fromYaml('''
input: int
output: "null"
pools:
  test:
    type:
      type: record
      name: MyRecord
      fields:
        - {name: a, type: int}
        - {name: b, type: string}
    shared: true
action:
  - pool: test
    path: [{string: zzz}, {string: a}]
    to:
      params: [{x: int}]
      ret: int
      do: {+: [x, input]}
    init: {value: {a: 0, b: hey}, type: MyRecord}
  - null
''')
        self.assertEqual(engine.snapshot().pools["test"].init, {})
        engine.action(3)
        self.assertEqual(json.loads(engine.snapshot().pools["test"].init["zzz"])["a"], 3)
        engine.action(2)
        self.assertEqual(json.loads(engine.snapshot().pools["test"].init["zzz"])["a"], 5)

    def testPublicPoolsUnion(self):
        engine, = PFAEngine.fromYaml('''
input: int
output: "null"
pools:
  test:
    type: [int, string]
    shared: true
action:
  - pool: test
    path: [{string: zzz}]
    to:
      params: [{x: [int, string]}]
      ret: [int, string]
      do:
        cast: x
        cases:
          - as: int
            named: y
            do: {+: [y, input]}
          - as: string
            named: y
            do: y
    init: {int: 0}
  - null
''')
        self.assertEqual(engine.snapshot().pools["test"].init, {})
        engine.action(3)
        self.assertEqual(engine.snapshot().pools["test"].init["zzz"], "3")
        engine.action(2)
        self.assertEqual(engine.snapshot().pools["test"].init["zzz"], "5")

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict
import copy
import json
import re
import sys

############################################## quick access functions

def splitIndex(index):
    """Normalize a casual index like "one, two, 3,four" to canonical form like ["one", "two", 3, "four"]."""
    out = [x.strip() for x in index.split(",")]
    for i in range(len(out)):
        try:
            out[i] = int(out[i])
        except ValueError:
            pass
    return out

def get(expr, index):
    """Get a subexpression from the root expr and index."""
    if isinstance(index, str):
        index = splitIndex(index)
    out = expr
    for i in index:
        out = out[i]
    return out

def assign(expr, index, to):
    """Destructively (in-place) replace a subexpression at a given index with another expression."""
    if isinstance(index, str):
        index = splitIndex(index)
    out = expr
    for i in index[:-1]:
        out = out[i]
    out[index[-1]] = to

def assigned(expr, index, to):
    """Return a copy of expr with a replaced subexpression, leaving the original intact."""
    expr = copy.deepcopy(expr)
    assign(expr, index, to)
    return expr

def assignAt(pattern, expr, to):
    """Destructively (in-place) replace subexpressions that match a regular expression pattern."""
    for i in indexes(pattern, expr):
        assign(expr, i, to)

def assignedAt(pattern, expr, to):
    """Return a copy of expr with replaced subexpressions at points that match a regular expression pattern."""
    expr = copy.deepcopy(expr)
    for i in indexes(pattern, expr):
        assign(expr, i, to)
    return expr

def remove(expr, index):
    """Destructively (in-place) remove a subexpression at a given index."""
    if isinstance(index, str):
        index = splitIndex(index)
    out = expr
    for i in index[:-1]:
        out = out[i]
    del out[index[-1]]

def removed(expr, index):
    """Return a copy of expr with a subexpression removed at a given index."""
    expr = copy.deepcopy(expr)
    remove(expr, index)
    return expr

############################################## pattern-matching primitives

class Match(object):
    """Represents an expression matched by a regular expression.

    original is a copy of the expression, identical to the original subexpression.
    modified is a copy of the expression with RegEx substitutions.
    groups is a dictionary of modified subexpressions."""

    def __init__(self, original, modified):
        self.original = original
        self.modified = modified
        self.groups = {}
    def __repr__(self):
        return "Match(" + repr(self.modified) + ")"

def getmatch(pattern, haystack):
    """Generic function to match a regular expression pattern to a haystack expression."""

    if isinstance(pattern, dict):
        if not isinstance(haystack, dict):
            return None
        if set(pattern.keys()).difference(set(["@"])) != set(haystack.keys()).difference(set(["@"])):
            return None
        out = Match({}, {})
        for key in pattern:
            if key != "@":
                m = getmatch(pattern[key], haystack[key])
                if m is not None:
                    out.original[key] = m.original
                    out.modified[key] = m.modified
                    out.groups.update(m.groups)
                else:
                    return None
        return out

    elif isinstance(pattern, (list, tuple)):
        if not isinstance(haystack, (list, tuple)):
            return None
        if len(pattern) != len(haystack):
            return None
        out = Match([], [])
        for index, item in enumerate(pattern):
            m = getmatch(item, haystack[index])
            if m is not None:
                out.original.append(m.original)
                out.modified.append(m.modified)
                out.groups.update(m.groups)
            else:
                return None
        return out

    elif isinstance(pattern, (str, int, float)):
        if pattern == haystack:
            return Match(haystack, haystack)
        else:
            return None

    elif pattern in (True, False, None):
        if pattern is haystack:
            return Match(haystack, haystack)
        else:
            return None

    else:
        return pattern.getmatch(haystack)

class Matcher(object):
    """Base class of regular expression patterns (other than literal JSON expressions)."""
    def getmatch(self, haystack):
        raise NotImplementedError

class Group(Matcher):
    """Tag a subexpression with a name so that it can be found later."""
    def __init__(self, **kwds):
        (self.name, self.matcher), = list(kwds.items())
    def __repr__(self):
        return "Group(" + repr(self.name) + "=" + repr(self.matcher) + ")"
    def getmatch(self, haystack):
        m = self.matcher.getmatch(haystack)
        if m is not None:
            m.groups[self.name] = m.modified
            return m
        else:
            return None

class Min(Matcher):
    """Match a dictionary that has at least the given key-value pairs."""
    def __init__(self, **obj):
        self.obj = obj
    def __repr__(self):
        return "Min(" + ", ".join(k + "=" + repr(v) for k, v in list(self.obj.items())) + ")"
    def getmatch(self, haystack):
        if not isinstance(haystack, dict):
            return None
        if len(set(self.obj.keys()).difference(set(["@"])).difference(set(haystack.keys()))) > 0:
            return None
        out = Match(dict((k, v) for k, v in list(haystack.items()) if k != "@"),
                    dict((k, v) for k, v in list(haystack.items()) if k != "@"))
        for key in self.obj:
            if key != "@":
                m = getmatch(self.obj[key], haystack[key])
                if m is not None:
                    out.original[key] = m.original
                    out.modified[key] = m.modified
                    out.groups.update(m.groups)
                else:
                    return None
        return out

class Start(Matcher):
    """Match a list that starts with the given elements."""
    def __init__(self, *array):
        self.array = array
    def __repr__(self):
        return "Start(" + ", ".join(map(repr, self.array)) + ")"
    def getmatch(self, haystack):
        if not isinstance(haystack, (list, tuple)):
            return None
        if len(self.array) > len(haystack):
            return None
        out = Match([], [])
        for index, item in enumerate(self.array):
            m = getmatch(item, haystack[index])
            if m is not None:
                out.original.append(m.original)
                out.modified.append(m.modified)
                out.groups.update(m.groups)
            else:
                return None
        out.original.extend(haystack[len(self.array):])
        out.modified.extend(haystack[len(self.array):])
        return out

class End(Matcher):
    """Match a list that ends with the given elements."""
    def __init__(self, *array):
        self.array = array
    def __repr__(self):
        return "End(" + ", ".join(map(repr, self.array)) + ")"
    def getmatch(self, haystack):
        if not isinstance(haystack, (list, tuple)):
            return None
        if len(self.array) > len(haystack):
            return None
        out = Match([], [])
        for index, item in enumerate(self.array):
            m = getmatch(item, haystack[len(haystack) - len(self.array) + index])
            if m is not None:
                out.original.append(m.original)
                out.modified.append(m.modified)
                out.groups.update(m.groups)
            else:
                return None
        out.original = haystack[:(len(haystack) - len(self.array))] + out.original
        out.modified = haystack[:(len(haystack) - len(self.array))] + out.modified
        return out

class Or(Matcher):
    """Match at least one of several alternatives.

    If any alternatives have substitutions, the first matching substitution is returned."""
    def __init__(self, *alternatives):
        self.alternatives = alternatives
    def __repr__(self):
        return "Or(" + ", ".join(repr(x) for x in self.alternatives) + ")"
    def getmatch(self, haystack):
        if len(self.alternatives) == 0:
            return None
        for alt in self.alternatives:
            m = getmatch(alt, haystack)
            if m is not None:
                return m
        return None

class And(Matcher):
    """Match every one of several alternatives.

    If any alternatives have substitutions, the last substitution is returned."""
    def __init__(self, *alternatives):
        self.alternatives = alternatives
    def __repr__(self):
        return "And(" + ", ".join(repr(x) for x in self.alternatives) + ")"
    def getmatch(self, haystack):
        if len(self.alternatives) == 0:
            return None
        for alt in self.alternatives:
            m = getmatch(alt, haystack)
            if m is None:
                return None
        return m

class Approx(Matcher):
    """Match a number within a given range."""
    def __init__(self, central, error):
        self.central = central
        self.error = error
    def __repr__(self):
        return "Approx(" + repr(self.central) + ", " + repr(self.error) + ")"
    def getmatch(self, haystack):
        if not isinstance(haystack, (int, float)) or abs(haystack - self.central) > self.error:
            return None
        else:
            return Match(haystack, haystack)

class LT(Matcher):
    """Match a number if it is less than some value."""
    def __init__(self, value):
        self.value = value
    def __repr__(self):
        return "LT(" + repr(self.value) + ")"
    def getmatch(self, haystack):
        if not isinstance(haystack, (int, float)) or haystack >= self.value:
            return None
        else:
            return Match(haystack, haystack)

class LE(Matcher):
    """Match a number if it is less than or equal to some value."""
    def __init__(self, value):
        self.value = value
    def __repr__(self):
        return "LE(" + repr(self.value) + ")"
    def getmatch(self, haystack):
        if not isinstance(haystack, (int, float)) or haystack > self.value:
            return None
        else:
            return Match(haystack, haystack)

class GT(Matcher):
    """Match a number if it is greater than some value."""
    def __init__(self, value):
        self.value = value
    def __repr__(self):
        return "GT(" + repr(self.value) + ")"
    def getmatch(self, haystack):
        if not isinstance(haystack, (int, float)) or haystack <= self.value:
            return None
        else:
            return Match(haystack, haystack)

class GE(Matcher):
    """Match a number if it is greater than or equal to some value."""
    def __init__(self, value):
        self.value = value
    def __repr__(self):
        return "GE(" + repr(self.value) + ")"
    def getmatch(self, haystack):
        if not isinstance(haystack, (int, float)) or haystack < self.value:
            return None
        else:
            return Match(haystack, haystack)

class RegEx(Matcher):
    """Match a string with a regular expression.

    If 'to' is given, this provides a substitution (using \\1, \\2, etc. variables).
    The 'flags' are iLmsux, corresponding to Python regular expression flags."""
    def __init__(self, pattern, to=None, flags=None):
        self.pattern = pattern
        self.to = to
        self.flags = flags
    def __repr__(self):
        return "RegEx(" + repr(self.pattern) + ", " + repr(self.to) + ", " + repr(self.flags) + ")"
    def getmatch(self, haystack):
        if not isinstance(haystack, str):
            return None
        flags = 0
        if self.flags is not None:
            if "i" in self.flags or "I" in self.flags:
                flags |= re.I
            if "l" in self.flags or "L" in self.flags:
                flags |= re.L
            if "m" in self.flags or "M" in self.flags:
                flags |= re.M
            if "s" in self.flags or "S" in self.flags:
                flags |= re.S
            if "u" in self.flags or "U" in self.flags:
                flags |= re.U
            if "x" in self.flags or "X" in self.flags:
                flags |= re.X
        if re.match(self.pattern, haystack, flags=flags) is None:
            return None
        elif self.to is None:
            return Match(haystack, haystack)
        else:
            return Match(haystack, re.sub(self.pattern, self.to, haystack, flags=flags))

class Any(Matcher):
    """Match any expression or any expression with one of several given types."""
    def __init__(self, *types):
        self.types = types
    def __repr__(self):
        return "Any(" + ", ".join(map(repr, self.types)) + ")"
    def getmatch(self, haystack):
        if len(self.types) == 0 or isinstance(haystack, self.types):
            return Match(haystack, haystack)
        else:
            return None

############################################## pattern-matching functions

def search(pattern, haystack, result=()):
    """Create a generator to yield matches as (index, Match) pairs."""
    if isinstance(haystack, dict):
        for key in sorted(haystack.keys()):
            for x in search(pattern, haystack[key], result + (key,)):
                yield x

    elif isinstance(haystack, (list, tuple)):
        for i, item in enumerate(haystack):
            for x in search(pattern, item, result + (i,)):
                yield x

    m = getmatch(pattern, haystack)
    if m is not None:
        yield result, m

def searchFirst(pattern, haystack):
    """Return a (index, Match) pair or None if there is no match."""
    for x in search(pattern, haystack):
        return x
    return None

def index(pattern, haystack):
    """Return an index or None if there is no match."""
    x = searchFirst(pattern, haystack)
    if x is None:
        return None
    else:
        return x[0]

def indexes(pattern, haystack):
    """Create a generator to yield matching indexes."""
    return (x[0] for x in search(pattern, haystack))

def contains(pattern, haystack):
    """Return True if the pattern is found; False otherwise."""
    return searchFirst(pattern, haystack) is not None

def count(pattern, haystack):
    """Return the number of matches."""
    return len(list(search(pattern, haystack)))

def findRef(pattern, haystack):
    """Return a reference to the matching subexpression within the original structure (for in-place modifications) or None if there is no match."""
    x = searchFirst(pattern, haystack)
    if x is None:
        return None
    else:
        return get(haystack, x[0])

def findCopy(pattern, haystack):
    """Return a copy of the matching subexpression or None if there is no match."""
    x = searchFirst(pattern, haystack)
    if x is None:
        return None
    else:
        return x[1].original   # "original" is a copy that is identical to the original (see Match class)

def findAllRef(pattern, haystack):
    """Create a generator to yield references to matching subexpressions within the original structure (for in-place modifications)."""
    for x in search(pattern, haystack):
        yield get(haystack, x[0])

def findAllCopy(pattern, haystack):
    """Create a generator to yield copies of the matching subexpressions."""
    for x in search(pattern, haystack):
        yield x[1].original   # "original" is a copy that is identical to the original (see Match class)

############################################## the look function presents a JSON object in an analyzable way

def look(expr, maxDepth=8, inlineDepth=2, indexWidth=30, dropAt=True, stream=sys.stdout):
    """Print a JSON object on the screen in a readable way.

    maxDepth: maximum depth to show before printing ellipsis (...)
    inlineDepth: maximum depth to show on a single line
    indexWidth: width (in characters) of the index column on the left
    dropAt: don't show "@" keys
    stream: allows the output to be sent to a file or stream.
    """
    if dropAt:
        expr = _dropAt(expr, maxDepth + inlineDepth)
    stream.write(("%%-%ds %%s\n" % indexWidth) % ("index", "data"))
    stream.write("-" * (indexWidth * 2) + "\n")
    for reprindex, reprdata in _look(expr, maxDepth, inlineDepth, [], indexWidth, ""):
        stream.write(("%%-%ds %%s\n" % indexWidth) % (reprindex, reprdata))
    stream.flush()

def _dropAt(expr, depth):
    if isinstance(expr, dict):
        return expr.__class__([(k, _dropAt(v, depth - 1)) for k, v in list(expr.items()) if k != "@"])
    elif isinstance(expr, (list, tuple)):
        return [_dropAt(x, depth - 1) for x in expr]
    else:
        return expr

def _acceptableDepth(expr, limit):
    if limit < 0:
        return False
    elif isinstance(expr, dict):
        return all(_acceptableDepth(x, limit - 1) for x in list(expr.values()))
    elif isinstance(expr, (list, tuple)):
        return all(_acceptableDepth(x, limit - 1) for x in expr)
    else:
        return True

class _Pair(object):
    def __init__(self, key, value):
        self.key = key
        self.value = value

class _Item(object):
    def __init__(self, value):
        self.value = value

def _look(expr, maxDepth, inlineDepth, index, indexWidth, indent):
    reprindex = ",".join(map(str, index))
    if len(reprindex) > indexWidth:
        reprindex = reprindex[:(indexWidth - 3)] + "..."

    if isinstance(expr, _Pair):
        block = _look(expr.value, maxDepth, inlineDepth, index, indexWidth, indent)
        block[0][0] = reprindex
        block[0][1] = indent + json.dumps(expr.key) + ": " + block[0][1].lstrip()
        block[-1][1] = block[-1][1] + ","
        return block

    elif isinstance(expr, _Item):
        block = _look(expr.value, maxDepth, inlineDepth, index, indexWidth, indent)
        block[0][0] = reprindex
        block[0][1] = indent + block[0][1].lstrip()
        block[-1][1] = block[-1][1] + ","
        return block

    elif _acceptableDepth(expr, inlineDepth):
        return [[reprindex, indent + json.dumps(expr)]]

    elif isinstance(expr, dict):
        if maxDepth > 0:
            block = []
            for key, value in list(expr.items()):
                block.extend(_look(_Pair(key, value), maxDepth - 1, inlineDepth, index + [key], indexWidth, indent + "  "))
            block[-1][1] = block[-1][1].rstrip(",")
            return [[reprindex, indent + "{"]] + block + [["", indent + "}"]]
        else:
            return [[reprindex, indent + "{...}"]]

    elif isinstance(expr, (list, tuple)):
        if maxDepth > 0:
            block = []
            for i, value in enumerate(expr):
                block.extend(_look(_Item(value), maxDepth - 1, inlineDepth, index + [i], indexWidth, indent + "  "))
            block[-1][1] = block[-1][1].rstrip(",")
            return [[reprindex, indent + "["]] + block + [["", indent + "]"]]
        else:
            return [[reprindex, indent + "[...]"]]



#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

from titus.fcn import Fcn
from titus.fcn import LibFcn
from titus.signature import Sig
from titus.signature import Sigs
from titus.datatype import *
from titus.errors import *
from titus.util import callfcn, div
import titus.P as P
from functools import reduce

provides = {}
def provide(fcn):
    provides[fcn.name] = fcn

prefix = "la."

def np():
    import numpy
    return numpy

def rowKeys(x):
    return set(x.keys())

def colKeys(x):
    if len(x) == 0:
        return set()
    else:
        return reduce(lambda a, b: a.union(b), [set(xi.keys()) for xi in list(x.values())])

def arraysToMatrix(x):
    return np().matrix(x, dtype=np().double)

def arrayToRowVector(x):
    return np().matrix(x, dtype=np().double).T

def rowVectorToArray(x):
    return x.T.tolist()[0]

def matrixToArrays(x):
    return x.tolist()

def mapsToMatrix(x, rows, cols):
    return np().matrix([[x.get(i, {}).get(j, 0.0) for j in cols] for i in rows], dtype=np().double)

def mapToRowVector(x, keys):
    return np().matrix([x.get(k, 0.0) for k in keys], dtype=np().double).T

def rowVectorToMap(x, keys):
    return dict(list(zip(keys, x.T.tolist()[0])))

def matrixToMaps(x, rows, cols):
    return dict((row, dict(list(zip(cols, xi)))) for row, xi in zip(rows, x.tolist()))

def raggedArray(x):
    collens = list(map(len, x))
    return max(collens) != min(collens)

def raggedMap(x):
    return len(set(len(xi) for xi in list(x.values()))) != 1

class MapApply(LibFcn):
    name = prefix + "map"
    sig = Sigs([Sig([{"x": P.Array(P.Array(P.Double()))}, {"fcn": P.Fcn([P.Double()], P.Double())}], P.Array(P.Array(P.Double()))),
                Sig([{"x": P.Map(P.Map(P.Double()))}, {"fcn": P.Fcn([P.Double()], P.Double())}], P.Map(P.Map(P.Double())))])
    errcodeBase = 24000
    def __call__(self, state, scope, pos, paramTypes, x, fcn):
        if isinstance(x, (list, tuple)) and all(isinstance(xi, (list, tuple)) for xi in x):
            return [[callfcn(state, scope, fcn, [xj]) for xj in xi] for xi in x]

        elif isinstance(x, dict) and all(isinstance(x[i], dict) for i in list(x.keys())):
            return dict((i, dict((j, callfcn(state, scope, fcn, [xj])) for j, xj in list(xi.items()))) for i, xi in list(x.items()))

provide(MapApply())

class Scale(LibFcn):
    name = prefix + "scale"
    sig = Sigs([Sig([{"x": P.Array(P.Double())}, {"alpha": P.Double()}], P.Array(P.Double())),
                Sig([{"x": P.Array(P.Array(P.Double()))}, {"alpha": P.Double()}], P.Array(P.Array(P.Double()))),
                Sig([{"x": P.Map(P.Double())}, {"alpha": P.Double()}], P.Map(P.Double())),
                Sig([{"x": P.Map(P.Map(P.Double()))}, {"alpha": P.Double()}], P.Map(P.Map(P.Double())))])
    errcodeBase = 24010
    def __call__(self, state, scope, pos, paramTypes, x, alpha):
        if isinstance(x, (list, tuple)) and all(isinstance(xi, (list, tuple)) for xi in x):
            return [[xj * alpha for xj in xi] for xi in x]
        elif isinstance(x, (list, tuple)):
            return [xi * alpha for xi in x]
        elif isinstance(x, dict) and all(isinstance(x[i], dict) for i in x):
            return dict((i, dict((j, xj * alpha) for j, xj in list(xi.items()))) for i, xi in list(x.items()))
        else:
            return dict((i, xi * alpha) for i, xi in list(x.items()))

provide(Scale())

class ZipMap(LibFcn):
    name = prefix + "zipmap"
    sig = Sigs([Sig([{"x": P.Array(P.Array(P.Double()))}, {"y": P.Array(P.Array(P.Double()))}, {"fcn": P.Fcn([P.Double(), P.Double()], P.Double())}], P.Array(P.Array(P.Double()))),
                Sig([{"x": P.Map(P.Map(P.Double()))}, {"y": P.Map(P.Map(P.Double()))}, {"fcn": P.Fcn([P.Double(), P.Double()], P.Double())}], P.Map(P.Map(P.Double())))])
    errcodeBase = 24020
    def __call__(self, state, scope, pos, paramTypes, x, y, fcn):
        if isinstance(x, (list, tuple)) and all(isinstance(xi, (list, tuple)) for xi in x) and \
           isinstance(y, (list, tuple)) and all(isinstance(yi, (list, tuple)) for yi in y):
            if len(x) != len(y) or any(len(xi) != len(yi) for xi, yi in zip(x, y)):
                raise PFARuntimeException("misaligned matrices", self.errcodeBase + 0, self.name, pos)
            return [[callfcn(state, scope, fcn, [xj, yj]) for xj, yj in zip(xi, yi)] for xi, yi in zip(x, y)]

        elif isinstance(x, dict) and all(isinstance(x[i], dict) for i in list(x.keys())) and \
             isinstance(y, dict) and all(isinstance(y[i], dict) for i in list(y.keys())):
            rows = rowKeys(x).union(rowKeys(y))
            cols = colKeys(x).union(colKeys(y))
            return dict((i, dict((j, callfcn(state, scope, fcn, [x.get(i, {}).get(j, 0.0), y.get(i, {}).get(j, 0.0)])) for j in cols)) for i in rows)

provide(ZipMap())

class Add(LibFcn):
    name = prefix + "add"
    sig = Sigs([Sig([{"x": P.Array(P.Double())}, {"y": P.Array(P.Double())}], P.Array(P.Double())),
                Sig([{"x": P.Array(P.Array(P.Double()))}, {"y": P.Array(P.Array(P.Double()))}], P.Array(P.Array(P.Double()))),
                Sig([{"x": P.Map(P.Double())}, {"y": P.Map(P.Double())}], P.Map(P.Double())),
                Sig([{"x": P.Map(P.Map(P.Double()))}, {"y": P.Map(P.Map(P.Double()))}], P.Map(P.Map(P.Double())))])
    errcodeBase = 24030
    def __call__(self, state, scope, pos, paramTypes, x, y):
        if isinstance(x, (list, tuple)) and all(isinstance(xi, (list, tuple)) for xi in x) and \
           isinstance(y, (list, tuple)) and all(isinstance(yi, (list, tuple)) for yi in y):
            if len(x) != len(y) or any(len(xi) != len(yi) for xi, yi in zip(x, y)):
                raise PFARuntimeException("misaligned matrices", self.errcodeBase + 0, self.name, pos)
            return [[xj + yj for xj, yj in zip(xi, yi)] for xi, yi in zip(x, y)]

        elif isinstance(x, (list, tuple)) and isinstance(y, (list, tuple)):
            if len(x) != len(y):
                raise PFARuntimeException("misaligned matrices", self.errcodeBase + 0, self.name, pos)
            return [xi + yi for xi, yi in zip(x, y)]

        elif isinstance(x, dict) and all(isinstance(x[i], dict) for i in list(x.keys())) and \
             isinstance(y, dict) and all(isinstance(y[i], dict) for i in list(y.keys())):
            rows = rowKeys(x).union(rowKeys(y))
            cols = colKeys(x).union(colKeys(y))
            return dict((i, dict((j, x.get(i, {}).get(j, 0.0) + y.get(i, {}).get(j, 0.0)) for j in cols)) for i in rows)

        else:
            rows = rowKeys(x).union(rowKeys(y))
            return dict((i, x.get(i, 0.0) + y.get(i, 0.0)) for i in rows)

provide(Add())

class Sub(LibFcn):
    name = prefix + "sub"
    sig = Sigs([Sig([{"x": P.Array(P.Double())}, {"y": P.Array(P.Double())}], P.Array(P.Double())),
                Sig([{"x": P.Array(P.Array(P.Double()))}, {"y": P.Array(P.Array(P.Double()))}], P.Array(P.Array(P.Double()))),
                Sig([{"x": P.Map(P.Double())}, {"y": P.Map(P.Double())}], P.Map(P.Double())),
                Sig([{"x": P.Map(P.Map(P.Double()))}, {"y": P.Map(P.Map(P.Double()))}], P.Map(P.Map(P.Double())))])
    errcodeBase = 24040
    def __call__(self, state, scope, pos, paramTypes, x, y):
        if isinstance(x, (list, tuple)) and all(isinstance(xi, (list, tuple)) for xi in x) and \
           isinstance(y, (list, tuple)) and all(isinstance(yi, (list, tuple)) for yi in y):
            if len(x) != len(y) or any(len(xi) != len(yi) for xi, yi in zip(x, y)):
                raise PFARuntimeException("misaligned matrices", self.errcodeBase + 0, self.name, pos)
            return [[xj - yj for xj, yj in zip(xi, yi)] for xi, yi in zip(x, y)]

        elif isinstance(x, (list, tuple)) and isinstance(y, (list, tuple)):
            if len(x) != len(y):
                raise PFARuntimeException("misaligned matrices", self.errcodeBase + 0, self.name, pos)
            return [xi - yi for xi, yi in zip(x, y)]

        elif isinstance(x, dict) and all(isinstance(x[i], dict) for i in list(x.keys())) and \
             isinstance(y, dict) and all(isinstance(y[i], dict) for i in list(y.keys())):
            rows = rowKeys(x).union(rowKeys(y))
            cols = colKeys(x).union(colKeys(y))
            return dict((i, dict((j, x.get(i, {}).get(j, 0.0) - y.get(i, {}).get(j, 0.0)) for j in cols)) for i in rows)

        else:
            rows = rowKeys(x).union(rowKeys(y))
            return dict((i, x.get(i, 0.0) - y.get(i, 0.0)) for i in rows)

provide(Sub())

class Dot(LibFcn):
    name = prefix + "dot"
    sig = Sigs([Sig([{"x": P.Array(P.Array(P.Double()))}, {"y": P.Array(P.Double())}], P.Array(P.Double())),
                Sig([{"x": P.Map(P.Map(P.Double()))}, {"y": P.Map(P.Double())}], P.Map(P.Double())),
                Sig([{"x": P.Array(P.Array(P.Double()))}, {"y": P.Array(P.Array(P.Double()))}], P.Array(P.Array(P.Double()))),
                Sig([{"x": P.Map(P.Map(P.Double()))}, {"y": P.Map(P.Map(P.Double()))}], P.Map(P.Map(P.Double())))])
    errcodeBase = 24050
    def __call__(self, state, scope, pos, paramTypes, x, y):
        if paramTypes[1]["type"] == "array":
            if isinstance(paramTypes[1]["items"], dict) and paramTypes[1]["items"]["type"] == "array":
                # array matrix-matrix case
                bad = any(any(math.isnan(z) or math.isinf(z) for z in row) for row in x) or \
                      any(any(math.isnan(z) or math.isinf(z) for z in row) for row in y)
                xmat = arraysToMatrix(x)
                ymat = arraysToMatrix(y)
                if xmat.shape[0] == 0 or xmat.shape[1] == 0 or ymat.shape[0] == 0 or ymat.shape[1] == 0:
                    raise PFARuntimeException("too few rows/cols", self.errcodeBase + 1, self.name, pos)
                try:
                    if bad: raise PFARuntimeException("contains non-finite value", self.errcodeBase + 2, self.name, pos)
                    return matrixToArrays(np().dot(xmat, ymat))
                except ValueError:
                    raise PFARuntimeException("misaligned matrices", self.errcodeBase + 0, self.name, pos)

            else:
                # array matrix-vector case
                bad = any(any(math.isnan(z) or math.isinf(z) for z in row) for row in x) or \
                      any(math.isnan(z) or math.isinf(z) for z in y)
                xmat = arraysToMatrix(x)
                ymat = arrayToRowVector(y)
                if xmat.shape[0] == 0 or xmat.shape[1] == 0 or ymat.shape[0] == 0 or ymat.shape[1] == 0:
                    raise PFARuntimeException("too few rows/cols", self.errcodeBase + 1, self.name, pos)
                try:
                    if bad: raise PFARuntimeException("contains non-finite value", self.errcodeBase + 2, self.name, pos)
                    return rowVectorToArray(np().dot(xmat, ymat))
                except ValueError:
                    raise PFARuntimeException("misaligned matrices", self.errcodeBase + 0, self.name, pos)

        elif paramTypes[1]["type"] == "map":
            if isinstance(paramTypes[1]["values"], dict) and paramTypes[1]["values"]["type"] == "map":
                # map matrix-matrix case
                bad = any(any(math.isnan(z) or math.isinf(z) for z in list(row.values())) for row in list(x.values())) or \
                      any(any(math.isnan(z) or math.isinf(z) for z in list(row.values())) for row in list(y.values()))
                rows = list(rowKeys(x))
                inter = list(colKeys(x).union(rowKeys(y)))
                cols = list(colKeys(y))
                xmat = mapsToMatrix(x, rows, inter)
                ymat = mapsToMatrix(y, inter, cols)
                if xmat.shape[0] == 0 or xmat.shape[1] == 0 or ymat.shape[0] == 0 or ymat.shape[1] == 0:
                    raise PFARuntimeException("too few rows/cols", self.errcodeBase + 1, self.name, pos)
                if bad: raise PFARuntimeException("contains non-finite value", self.errcodeBase + 2, self.name, pos)
                return matrixToMaps(np().dot(xmat, ymat), rows, cols)

            else:
                # map matrix-vector case
                bad = any(any(math.isnan(z) or math.isinf(z) for z in list(row.values())) for row in list(x.values())) or \
                      any(math.isnan(z) or math.isinf(z) for z in list(y.values()))
                rows = list(rowKeys(x))
                cols = list(colKeys(x).union(rowKeys(y)))
                xmat = mapsToMatrix(x, rows, cols)
                ymat = mapToRowVector(y, cols)
                if xmat.shape[0] == 0 or xmat.shape[1] == 0 or ymat.shape[0] == 0 or ymat.shape[1] == 0:
                    raise PFARuntimeException("too few rows/cols", self.errcodeBase + 1, self.name, pos)
                if bad: raise PFARuntimeException("contains non-finite value", self.errcodeBase + 2, self.name, pos)
                return rowVectorToMap(np().dot(xmat, ymat), rows)

provide(Dot())
    
class Transpose(LibFcn):
    name = prefix + "transpose"
    sig = Sigs([Sig([{"x": P.Array(P.Array(P.Double()))}], P.Array(P.Array(P.Double()))),
                Sig([{"x": P.Map(P.Map(P.Double()))}], P.Map(P.Map(P.Double())))])
    errcodeBase = 24060
    def __call__(self, state, scope, pos, paramTypes, x):
        if isinstance(x, (list, tuple)) and all(isinstance(xi, (list, tuple)) for xi in x):
            rows = len(x)
            if rows < 1:
                raise PFARuntimeException("too few rows/cols", self.errcodeBase + 0, self.name, pos)
            cols = len(x[0])
            if cols < 1:
                raise PFARuntimeException("too few rows/cols", self.errcodeBase + 0, self.name, pos)
            if raggedArray(x):
                raise PFARuntimeException("ragged columns", self.errcodeBase + 1, self.name, pos)
            return [[x[r][c] for r in range(rows)] for c in range(cols)]

        elif isinstance(x, dict) and all(isinstance(x[i], dict) for i in list(x.keys())):
            rows = rowKeys(x)
            cols = colKeys(x)
            if len(rows) < 1 or len(cols) < 1:
                raise PFARuntimeException("too few rows/cols", self.errcodeBase + 0, self.name, pos)
            if raggedMap(x):
                raise PFARuntimeException("ragged columns", self.errcodeBase + 1, self.name, pos)
            return dict((c, dict((r, x[r][c]) for r in rows)) for c in cols)

provide(Transpose())

class Inverse(LibFcn):
    name = prefix + "inverse"
    sig = Sigs([Sig([{"x": P.Array(P.Array(P.Double()))}], P.Array(P.Array(P.Double()))),
                Sig([{"x": P.Map(P.Map(P.Double()))}], P.Map(P.Map(P.Double())))])
    errcodeBase = 24070
    def __call__(self, state, scope, pos, paramTypes, x):
        if isinstance(x, (list, tuple)) and all(isinstance(xi, (list, tuple)) for xi in x):
            rows = len(x)
            if rows < 1:
                raise PFARuntimeException("too few rows/cols", self.errcodeBase + 0, self.name, pos)
            cols = len(x[0])
            if cols < 1:
                raise PFARuntimeException("too few rows/cols", self.errcodeBase + 0, self.name, pos)
            if raggedArray(x):
                raise PFARuntimeException("ragged columns", self.errcodeBase + 1, self.name, pos)
            return matrixToArrays(arraysToMatrix(x).I)

        elif isinstance(x, dict) and all(isinstance(x[i], dict) for i in list(x.keys())):
            rows = list(rowKeys(x))
            cols = list(colKeys(x))
            if len(rows) < 1 or len(cols) < 1:
                raise PFARuntimeException("too few rows/cols", self.errcodeBase + 0, self.name, pos)
            xmat = mapsToMatrix(x, rows, cols)
            return matrixToMaps(xmat.I, cols, rows)

provide(Inverse())

class Trace(LibFcn):
    name = prefix + "trace"
    sig = Sigs([Sig([{"x": P.Array(P.Array(P.Double()))}], P.Double()),
                Sig([{"x": P.Map(P.Map(P.Double()))}], P.Double())])
    errcodeBase = 24080
    def __call__(self, state, scope, pos, paramTypes, x):
        if isinstance(x, (list, tuple)) and all(isinstance(xi, (list, tuple)) for xi in x):
            rows = len(x)
            if rows == 0:
                return 0.0
            else:
                cols = len(x[0])
                if raggedArray(x):
                    raise PFARuntimeException("ragged columns", self.errcodeBase + 0, self.name, pos)
                return sum(x[i][i] for i in range(min(rows, cols)))

        elif isinstance(x, dict) and all(isinstance(x[i], dict) for i in list(x.keys())):
            keys = rowKeys(x).intersection(colKeys(x))
            return sum(x[i][i] for i in keys)

provide(Trace())

class Det(LibFcn):
    name = prefix + "det"
    sig = Sigs([Sig([{"x": P.Array(P.Array(P.Double()))}], P.Double()),
                Sig([{"x": P.Map(P.Map(P.Double()))}], P.Double())])
    errcodeBase = 24090
    def __call__(self, state, scope, pos, paramTypes, x):
        if isinstance(x, (list, tuple)) and all(isinstance(xi, (list, tuple)) for xi in x):
            rows = len(x)
            if rows < 1:
                raise PFARuntimeException("too few rows/cols", self.errcodeBase + 0, self.name, pos)
            cols = len(x[0])
            if cols < 1:
                raise PFARuntimeException("too few rows/cols", self.errcodeBase + 0, self.name, pos)
            if raggedArray(x):
                raise PFARuntimeException("ragged columns", self.errcodeBase + 1, self.name, pos)
            if rows != cols:
                raise PFARuntimeException("non-square matrix", self.errcodeBase + 2, self.name, pos)
            if any(any(math.isnan(z) or math.isinf(z) for z in row) for row in x):
                return float("nan")
            else:
                return float(np().linalg.det(arraysToMatrix(x)))

        elif isinstance(x, dict) and all(isinstance(x[i], dict) for i in list(x.keys())):
            keys = list(rowKeys(x).union(colKeys(x)))
            if len(keys) < 1 or all(len(row) == 0 for row in list(x.values())):
                raise PFARuntimeException("too few rows/cols", self.errcodeBase + 0, self.name, pos)
            if any(any(math.isnan(z) or math.isinf(z) for z in list(row.values())) for row in list(x.values())):
                return float("nan")
            else:
                return float(np().linalg.det(mapsToMatrix(x, keys, keys)))

provide(Det())

class Symmetric(LibFcn):
    name = prefix + "symmetric"
    sig = Sigs([Sig([{"x": P.Array(P.Array(P.Double()))}, {"tol": P.Double()}], P.Boolean()),
                Sig([{"x": P.Map(P.Map(P.Double()))}, {"tol": P.Double()}], P.Boolean())])
    errcodeBase = 24100
    @staticmethod
    def same(x, y, tol):
        if math.isinf(x) and math.isinf(y) and ((x > 0.0 and y > 0.0) or (x < 0.0 and y < 0.0)):
            return True
        elif math.isnan(x) and math.isnan(y):
            return True
        elif not math.isinf(x) and not math.isnan(x) and not math.isinf(y) and not math.isnan(y):
            return abs(x - y) < tol
        else:
            return False
    def __call__(self, state, scope, pos, paramTypes, x, tol):
        if isinstance(x, (list, tuple)) and all(isinstance(xi, (list, tuple)) for xi in x):
            rows = len(x)
            if rows < 1:
                raise PFARuntimeException("too few rows/cols", self.errcodeBase + 0, self.name, pos)
            cols = len(x[0])
            if cols < 1:
                raise PFARuntimeException("too few rows/cols", self.errcodeBase + 0, self.name, pos)
            if raggedArray(x):
                raise PFARuntimeException("ragged columns", self.errcodeBase + 1, self.name, pos)
            if rows != cols:
                raise PFARuntimeException("non-square matrix", self.errcodeBase + 2, self.name, pos)
            return all(all(self.same(x[i][j], x[j][i], tol) for j in range(cols)) for i in range(rows))

        elif isinstance(x, dict) and all(isinstance(x[i], dict) for i in list(x.keys())):
            keys = list(rowKeys(x).union(colKeys(x)))
            if len(keys) < 1 or all(len(row) == 0 for row in list(x.values())):
                raise PFARuntimeException("too few rows/cols", self.errcodeBase + 0, self.name, pos)
            return all(all(self.same(x.get(i, {}).get(j, 0.0), x.get(j, {}).get(i, 0.0), tol) for j in keys) for i in keys)

provide(Symmetric())

class EigenBasis(LibFcn):
    name = prefix + "eigenBasis"
    sig = Sigs([Sig([{"x": P.Array(P.Array(P.Double()))}], P.Array(P.Array(P.Double()))),
                Sig([{"x": P.Map(P.Map(P.Double()))}], P.Map(P.Map(P.Double())))])

    errcodeBase = 24110
    def calculate(self, x, size):
        symm = (x + x.T) * 0.5

        evals, evects = np().linalg.eig(symm)
        evects = np().array(evects)
        evects2 = [evects[:,i] * (-1.0 if evects[0,i] < 0.0 else 1.0) for i in range(size)]

        eigvalm2 = [div(1.0, math.sqrt(abs(ei))) for ei in evals]
        order = np().argsort(eigvalm2)

        out = np().empty((size, size), dtype=np().double)
        for i in range(size):
            for j in range(size):
                out[i,j] = evects2[order[i]][j] * eigvalm2[order[i]]
        return out

    def __call__(self, state, scope, pos, paramTypes, x):
        if isinstance(x, (list, tuple)) and all(isinstance(xi, (list, tuple)) for xi in x):
            rows = len(x)
            if rows < 1:
                raise PFARuntimeException("too few rows/cols", self.errcodeBase + 0, self.name, pos)
            cols = len(x[0])
            if cols < 1:
                raise PFARuntimeException("too few rows/cols", self.errcodeBase + 0, self.name, pos)
            if raggedArray(x):
                raise PFARuntimeException("ragged columns", self.errcodeBase + 1, self.name, pos)
            if rows != cols:
                raise PFARuntimeException("non-square matrix", self.errcodeBase + 2, self.name, pos)
            if any(any(math.isnan(z) or math.isinf(z) for z in row) for row in x):
                raise PFARuntimeException("non-finite matrix", self.errcodeBase + 3, self.name, pos)
            return matrixToArrays(self.calculate(arraysToMatrix(x), rows))

        elif isinstance(x, dict) and all(isinstance(x[i], dict) for i in list(x.keys())):
            keys = list(rowKeys(x).union(colKeys(x)))
            if len(keys) < 1 or all(len(z) == 0 for z in list(x.values())):
                raise PFARuntimeException("too few rows/cols", self.errcodeBase + 0, self.name, pos)
            if any(any(math.isnan(z) or math.isinf(z) for z in list(row.values())) for row in list(x.values())):
                raise PFARuntimeException("non-finite matrix", self.errcodeBase + 3, self.name, pos)
            return matrixToMaps(self.calculate(mapsToMatrix(x, keys, keys), len(keys)), list(map(str, range(len(keys)))), keys)

provide(EigenBasis())

class Truncate(LibFcn):
    name = prefix + "truncate"
    sig = Sigs([Sig([{"x": P.Array(P.Array(P.Double()))}, {"keep": P.Int()}], P.Array(P.Array(P.Double()))),
                Sig([{"x": P.Map(P.Map(P.Double()))}, {"keep": P.Array(P.String())}], P.Map(P.Map(P.Double())))])
    errcodeBase = 24120
    def __call__(self, state, scope, pos, paramTypes, x, keep):
        if isinstance(keep, int) and keep < 0:
            keep = 0

        if isinstance(x, (list, tuple)) and all(isinstance(xi, (list, tuple)) for xi in x):
            rows = len(x)
            if rows < 1:
                raise PFARuntimeException("too few rows/cols", self.errcodeBase + 0, self.name, pos)
            cols = len(x[0])
            if cols < 1:
                raise PFARuntimeException("too few rows/cols", self.errcodeBase + 0, self.name, pos)
            if raggedArray(x):
                raise PFARuntimeException("ragged columns", self.errcodeBase + 1, self.name, pos)
            return x[:keep]

        elif isinstance(x, dict) and all(isinstance(x[i], dict) for i in list(x.keys())):
            rows = rowKeys(x)
            cols = colKeys(x)
            if len(rows) < 1 or len(cols) < 1:
                raise PFARuntimeException("too few rows/cols", self.errcodeBase + 0, self.name, pos)
            return dict((k, x[k]) for k in rows if k in keep)

provide(Truncate())



#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import math

import avro.io
import avro.schema

import titus.errors
import titus.util
from titus.util import ts, stringToBytes

######################################################### the most general types

class Type(object):
    """Superclass of all Avro types and also inline functions, which can only appear in argument lists."""
    @property
    def avroType(self): raise TypeError

class FcnType(Type):
    """Pseudo-type for inlines functions that can appear in argument lists."""

    def __init__(self, params, ret):
        self._params = params
        self._ret = ret

    @property
    def params(self):
        return self._params

    @property
    def ret(self):
        return self._ret

    def accepts(self, other):
        if isinstance(self, FcnType):
            return len(self.params) == len(other.params) and \
                   all(y.accepts(x) for x, y in zip(self.params, other.params)) and \
                   self.ret.accepts(other.ret)
        else:
            return False

    def __eq__(self, other):
        if isinstance(other, FcnType):
            return self.params == other.params and self.ret == other.ret
        else:
            return False

    def __repr__(self):
        return '{{"type": "function", "params": [{params}], "ret": {ret}}}'.format(
            params=",".join(repr(x) for x in self.params),
            ret=repr(self.ret))

    def jsonNode(self, memo):
        return {"type": "function", "params": [x.jsonNode(memo) for x in self.params], "ret": self.ret.jsonNode(memo)}

######################################################### Avro types

def jsonToAvroType(x):
    """Convert a JSON-encoded Avro type (AVSC file) into a titus.datatype.AvroType.

    :type x: string
    :param x: JSON-encoded Avro type
    :rtype: titus.datatype.AvroType
    :return: AvroType object
    """

    return schemaToAvroType(avro.schema.Parse(x))

def jsonNodeToAvroType(x):
    """Convert a Python-encoded Avro type into a titus.datatype.AvroType.

    :type x: dicts, lists, strings, numbers, ``True``, ``False``, ``None``
    :param x: Avro type in Python form
    :rtype: titus.datatype.AvroType
    :return: AvroType object
    """

    return schemaToAvroType(avro.schema.Parse(json.dumps(x)))

def schemaToAvroType(schema):
    """Convert an Avro schema into a titus.datatype.AvroType.

    :type schema: avro.schema.Schema
    :param schema: schema object from the Avro library
    :rtype: titus.datatype.AvroType
    :return: AvroType object
    """

    if schema.type == "null":
        return AvroNull()
    elif schema.type == "boolean":
        return AvroBoolean()
    elif schema.type == "int":
        return AvroInt()
    elif schema.type == "long":
        return AvroLong()
    elif schema.type == "float":
        return AvroFloat()
    elif schema.type == "double":
        return AvroDouble()
    elif schema.type == "bytes":
        return AvroBytes()
    elif schema.type == "fixed":
        out = AvroFixed.__new__(AvroFixed)
        out._schema = schema
        return out
    elif schema.type == "string":
        return AvroString()
    elif schema.type == "enum":
        out = AvroEnum.__new__(AvroEnum)
        out._schema = schema
        return out
    elif schema.type == "array":
        out = AvroArray.__new__(AvroArray)
        out._schema = schema
        return out
    elif schema.type == "map":
        out = AvroMap.__new__(AvroMap)
        out._schema = schema
        return out
    elif schema.type == "record":
        out = AvroRecord.__new__(AvroRecord)
        out._schema = schema
        return out
    elif schema.type == "union":
        out = AvroUnion.__new__(AvroUnion)
        out._schema = schema
        return out

def avroTypeToSchema(avroType):
    """Convert a titus.datatype.AvroType into an Avro schema.

    :type avroType: titus.datatype.AvroType
    :param avroType: AvroType object
    :rtype: avro.schema.Schema
    :return: schema object for the Avro library
    """

    return avroType.schema

class AvroType(Type):
    """Base class for types of all PFA/Avro values.

    Thin wrapper around a avro.schema.Schema, providing a different Python class for each Avro kind.
    """

    @property
    def schema(self):
        """Return the avro.schema.Schema object."""
        return self._schema

    @property
    def name(self):
        """Return "name" of this type, which is used as a key in tagged unions."""
        return None

    def __eq__(self, other):
        """Return ``True`` if the two types are equal."""
        if isinstance(other, AvroType):
            return self.schema == other.schema
        elif isinstance(other, AvroPlaceholder):
            return self.schema == other.avroType.schema
        else:
            return False

    def __hash__(self):
        """Return a unique hash of the type."""
        return hash(str(self.schema))

    def _recordFieldsOkay(self, other, memo, checkRecord):
        for xf in self.fields:
            if xf.default is None:
                if not any(xf.name == yf.name and xf.avroType.accepts(yf.avroType, memo, checkRecord) for yf in other.fields):
                    return False
            else:
                # not having a matching name in y is fine: x has a default
                # but having a matching name with a mismatched type is bad
                # (spec isn't clear, but org.apache.avro.SchemaCompatibility works that way)
                for yf in other.fields:
                    if xf.name == yf.name:
                        if not xf.avroType.accepts(yf.avroType, memo, checkRecord):
                            return False
        return True

    # self == "reader" (the anticipated signature, pattern to be matched),
    # other == "writer" (the given fact, argument to be accepted or rejected)
    # the special cases handle situations in which Python-Avro fails to be fully covariant
    def accepts(self, other, memo=None, checkRecord=True):
        """Perform type resolution: would PFA objects of type ``other`` be accepted where a PFA object of type ``self`` is expected?

        If ``x.accepts(y) and y.accepts(x)``, then ``x`` and ``y`` are equal. In general, acceptability is not symmetric.

        The ``memo`` and ``checkRecord`` parameters are only used to avoid infinite recursion.

        :type self: titus.datatype.AvroType
        :param self: the expected signature to be matched
        :type other: titus.datatype.AvroType
        :param other: the given argument to be accepted or rejected
        :rtype: boolean
        :return: ``True`` if ``other`` is an acceptable substitute for ``self`` (or is exactly the same); ``False`` if incompatible
        """

        if isinstance(other, ExceptionType):
            return False

        elif isinstance(self, AvroNull) and isinstance(other, AvroNull):
            return True
        elif isinstance(self, AvroBoolean) and isinstance(other, AvroBoolean):
            return True
        elif isinstance(self, AvroBytes) and isinstance(other, AvroBytes):
            return True
        elif isinstance(self, AvroString) and isinstance(other, AvroString):
            return True

        elif isinstance(self, AvroInt) and isinstance(other, AvroInt):
            return True
        elif isinstance(self, AvroLong) and (isinstance(other, AvroInt) or isinstance(other, AvroLong)):
            return True
        elif isinstance(self, AvroFloat) and (isinstance(other, AvroInt) or isinstance(other, AvroLong) or isinstance(other, AvroFloat)):
            return True
        elif isinstance(self, AvroDouble) and (isinstance(other, AvroInt) or isinstance(other, AvroLong) or isinstance(other, AvroFloat) or isinstance(other, AvroDouble)):
            return True

        elif isinstance(self, AvroArray) and isinstance(other, AvroArray):
            return self.items.accepts(other.items, memo, checkRecord)

        elif isinstance(self, AvroMap) and isinstance(other, AvroMap):
            return self.values.accepts(other.values, memo, checkRecord)

        elif isinstance(self, AvroFixed) and isinstance(other, AvroFixed):
            return self.size == other.size and self.fullName == other.fullName

        elif isinstance(self, AvroEnum) and isinstance(other, AvroEnum):
            return set(other.symbols).issubset(set(self.symbols)) and self.fullName == other.fullName

        elif isinstance(self, AvroRecord) and isinstance(other, AvroRecord):
            if memo is None:
                memo = set()
            else:
                memo = set(memo)

            if self.fullName != other.fullName:
                return False

            elif checkRecord and other.fullName not in memo:
                if not self._recordFieldsOkay(other, memo, checkRecord=False):
                    return False

                memo.add(self.fullName)

                if not self._recordFieldsOkay(other, memo, checkRecord):
                    return False

            return True

        elif isinstance(self, AvroUnion) and isinstance(other, AvroUnion):
            for yt in other.types:
                if not any(xt.accepts(yt, memo, checkRecord) for xt in self.types):
                    return False
            return True

        elif isinstance(self, AvroUnion):
            return any(xt.accepts(other, memo, checkRecord) for xt in self.types)

        elif isinstance(other, AvroUnion):
            return all(self.accepts(yt, memo, checkRecord) for yt in other.types)

        else:
            return False

    def toJson(self):
        """Convert the type to a JSON-encoded string, suitable for an AVSC file."""
        return json.dumps(self.jsonNode(set()))

    def jsonNode(self, memo):
        """Convert the type to a Python-encoded Avro type."""
        return self.name

    @property
    def avroType(self): return self

    def __repr__(self):
        return str(self.schema)

class AvroCompiled(AvroType):
    """AvroTypes that would be compiled in Java (AvroRecord, AvroFixed, AvroEnum)."""
    @property
    def name(self):
        """Name without namespace; unions would not use this as a tag; they would use fullName."""
        return self.schema.name
    @property
    def namespace(self):
        """Optional namespace; may be ``None``."""
        return self.schema.namespace
    @property
    def fullName(self):
        """Optional namespace and name joined with a dot."""
        if self.schema.namespace is None or self.schema.namespace == '':
            return self.schema.name
        return self.schema.fullname

class AvroNumber(AvroType):
    """Numeric AvroTypes (AvroInt, AvroLong, AvroFloat, AvroDouble)."""
    pass
class AvroRaw(AvroType):
    """Raw-byte AvroTypes (AvroBytes, AvroFixed)."""
    pass
class AvroIdentifier(AvroType):
    """AvroTypes that can be used as identifiers (AvroString, AvroEnum)."""
    pass
class AvroContainer(AvroType):
    """AvroTypes that contain other AvroTypes (AvroArray, AvroMap, AvroRecord)."""
    pass
class AvroMapping(AvroType):
    """AvroTypes that are represented by a JSON object in JSON (AvroMap, AvroRecord)."""
    pass

class ExceptionType(AvroType):
    """Pseudo-type for exceptions (the "bottom" type in type theory)."""
    def accepts(self, other):
        return isinstance(other, ExceptionType)
    def __repr__(self):
        return '{"type":"exception"}'
    def jsonNode(self, memo):
        return {"type": "exception"}
    @property
    def schema(self):
        return AvroNull().schema

######################################################### Avro type wrappers

class AvroNull(AvroType):
    """Avro "null" type. Has only one possible value, ``null``."""
    _schema = avro.schema.PrimitiveSchema("null")
    @property
    def name(self):
        return "null"

class AvroBoolean(AvroType):
    """Avro "boolean" type. Has only two possible values, ``true`` and ``false``."""
    _schema = avro.schema.PrimitiveSchema("boolean")
    @property
    def name(self):
        return "boolean"

class AvroInt(AvroNumber):
    """Avro "int" type for 32-bit integers."""
    _schema = avro.schema.PrimitiveSchema("int")
    @property
    def name(self):
        return "int"

class AvroLong(AvroNumber):
    """Avro "long" type for 64-bit integers."""
    _schema = avro.schema.PrimitiveSchema("long")
    @property
    def name(self):
        return "long"

class AvroFloat(AvroNumber):
    """Avro "float" type for 32-bit IEEE floating-point numbers."""
    _schema = avro.schema.PrimitiveSchema("float")
    @property
    def name(self):
        return "float"

class AvroDouble(AvroNumber):
    """Avro "double" type for 64-bit IEEE floating-point numbers."""
    _schema = avro.schema.PrimitiveSchema("double")
    @property
    def name(self):
        return "double"

class AvroBytes(AvroRaw):
    """Avro "bytes" type for arbitrary byte arrays."""
    _schema = avro.schema.PrimitiveSchema("bytes")
    @property
    def name(self):
        return "bytes"

class AvroFixed(AvroRaw, AvroCompiled):
    """Avro "fixed" type for fixed-length byte arrays."""
    def __init__(self, size, name=None, namespace=None):
        """Create an AvroFixed manually.

        :type size: int
        :param size: length of the fixed-length byte arrays
        :type name: string or ``NoneType``
        :param name: name or an auto-generated name
        :type namespace: string or ``NoneType``
        :param namespace: namespace or no namespace
        """

        if name is None:
            name = titus.util.uniqueFixedName()
        self._schema = avro.schema.FixedSchema(name, namespace, size, avro.schema.Names())
    @property
    def size(self):
        """Length of the fixed-length byte arrays."""
        return self.schema.size
    def jsonNode(self, memo):
        if self.fullName in memo:
            return self.fullName
        else:
            memo.add(self.fullName)
            out = {"type": "fixed", "size": self.size}
            if self.namespace is not None and self.namespace != "":
                out["name"] = self.name
                out["namespace"] = self.namespace
            else:
                out["name"] = self.name
            return out

class AvroString(AvroIdentifier): 
    """Avro "string" type for UTF-8 encoded strings."""
    _schema = avro.schema.PrimitiveSchema("string")
    @property
    def name(self):
        return "string"

class AvroEnum(AvroIdentifier, AvroCompiled):
    """Avro "enum" type for a small collection of string-labeled values."""
    def __init__(self, symbols, name=None, namespace=None):
        """Create an AvroEnum manually.

        :type symbols: list of strings
        :param symbols: collection of labels
        :type name: string or ``NoneType``
        :param name: name or an auto-generated name
        :type namespace: string or ``NoneType``
        :param namespace: namespace or no namespace
        """
        if name is None:
            name = titus.util.uniqueEnumName()
        self._schema = avro.schema.EnumSchema(name, namespace, symbols, avro.schema.Names())
    @property
    def symbols(self):
        """Allowed values for instances of this type."""
        return self.schema.symbols
    def jsonNode(self, memo):
        if self.fullName in memo:
            return self.fullName
        else:
            memo.add(self.fullName)
            out = {"type": "enum", "symbols": self.symbols}
            if self.namespace is not None and self.namespace != "":
                out["name"] = self.name
                out["namespace"] = self.namespace
            else:
                out["name"] = self.name
            return out

class AvroArray(AvroContainer):
    """Avro "array" type for homogeneous lists."""
    def __init__(self, items):
        """Create an AvroArray manually.

        :type items: titus.datatype.AvroType
        :param items: type of the contained objects
        """
        self._schema = avro.schema.ArraySchema(items.schema)
    @property
    def items(self):
        """Type of the contained objects."""
        return schemaToAvroType(self.schema.items)
    @property
    def name(self):
        return "array"
    def jsonNode(self, memo):
        return {"type": "array", "items": self.items.jsonNode(memo)}

class AvroMap(AvroContainer, AvroMapping):
    """Avro "map" type for homogeneous maps (keys must be strings)."""
    def __init__(self, values):
        """Create an AvroMap manually.

        :type values: titus.datatype.AvroType
        :param values: type of the contained objects
        """
        self._schema = avro.schema.MapSchema(values.schema)
    @property
    def values(self):
        """Type of the contained objects."""
        return schemaToAvroType(self.schema.values)
    @property
    def name(self):
        return "map"
    def jsonNode(self, memo):
        return {"type": "map", "values": self.values.jsonNode(memo)}

class AvroRecord(AvroContainer, AvroMapping, AvroCompiled):
    """Avro "record" type for inhomogeneous collections of named (and required) fields."""
    def __init__(self, fields, name=None, namespace=None):
        """Create an AvroRecord manually.

        :type fields: list of titus.datatype.AvroField
        :param fields: field names and types in order
        :type name: string or ``NoneType``
        :param name: name or an auto-generated name
        :type namespace: string or ``NoneType``
        :param namespace: namespace or no namespace
        """
        if name is None:
            name = titus.util.uniqueRecordName()
        self._schema = avro.schema.RecordSchema(name, namespace, [x.schema for x in fields], None, avro.schema.Names(), "record")
    @property
    def fields(self):
        """Get the fields as a list of titus.datatype.AvroField objects."""
        return [AvroField.fromSchema(x) for x in self.schema.fields]
    @property
    def fieldsDict(self):
        """Get the fields as a dict from field names to titus.datatype.AvroField objects."""
        return dict((x.name, x) for x in self.fields)
    def field(self, name):
        """Get one titus.datatype.AvroField object by name."""
        return self.fieldsDict[name]
    def jsonNode(self, memo):
        if self.fullName in memo:
            return self.fullName
        else:
            memo.add(self.fullName)
            out = {"type": "record"}
            if self.namespace is not None and self.namespace != "":
                out["name"] = self.name
                out["namespace"] = self.namespace
            else:
                out["name"] = self.name
            out["fields"] = []
            for field in self.fields:
                out["fields"].append(field.jsonNode(memo))
            return out

class AvroUnion(AvroType):
    """Avro "union" type for tagged unions."""
    def __init__(self, types):
        """Create an AvroUnion manually.

        :type types: list of titus.datatype.AvroType
        :param types: possible types for this union in the order of their resolution
        """
        names = set([x.name for x in types])
        if len(types) != len(names):
            raise titus.errors.AvroException("duplicate in union: " + ", ".join(map(ts, types)))
        if "union" in names:
            raise titus.errors.AvroException("nested union: " + ", ".join(map(ts, types)))
        self._schema = avro.schema.UnionSchema([x.schema for x in types])
    @property
    def types(self):
        """Get the possible types for this union as a list of titus.datatype.AvroType."""
        return [schemaToAvroType(x) for x in self._schema._schemas]
    @property
    def name(self):
        return "union"
    def jsonNode(self, memo):
        return [x.jsonNode(memo) for x in self.types]

class AvroField(object):
    """Field of a titus.datatype.AvroRecord."""
    @staticmethod
    def fromSchema(schema):
        """Create a field from an avro.schema.Schema."""
        out = AvroField.__new__(AvroField)
        out._schema = schema
        return out
    def __init__(self, name, avroType, index=None, default=None, order=None):
        """Create an AvroField manually.

        :type name: string
        :param name: field name
        :type avroType: titus.datatype.AvroType
        :param avroType: field type
        :type default: value or ``NoneType``
        :param default: default value or ``None``
        :type order: "ascending", "descending", or "ignore"
        :param order: sort order (used in Hadoop secondary sort)
        """
        self._schema = avro.schema.Field(avroType.schema, name, index, default is not None, default, order, avro.schema.Names())
    @property
    def schema(self):
        """Get the field type as an avro.schema.Schema."""
        return self._schema
    def __repr__(self):
        return json.dumps(self.schema.to_json())
    @property
    def name(self):
        """Get the field name."""
        return self.schema.name
    @property
    def avroType(self):
        """Get the field type as a titus.datatype.AvroType."""
        return schemaToAvroType(self.schema.type)
    @property
    def default(self):
        """Get the field default value."""
        if self.schema.has_default:
            return self.schema.default
        else:
            return None
    @property
    def order(self):
        """Get the field order."""
        return self.schema.order
    def jsonNode(self, memo):
        """Get the field as a Python-encoded JSON node."""
        out = {"name": self.name, "type": self.avroType.jsonNode(memo)}
        if self.default is not None:
            out["default"] = self.default
        if self.order is not None:
            out["order"] = self.order
        return out

########################### resolving types out of order in streaming input

class AvroPlaceholder(object):
    """Represents a type that can't be resolved yet because JSON objects may be streamed in an unknown order."""

    def __init__(self, original, forwardDeclarationParser):
        self.original = original
        self.forwardDeclarationParser = forwardDeclarationParser
        
    @property
    def avroType(self):
        """Called after ``titus.datatype.AvroTypeBuilder`` ``resolveTypes`` to get the resolved type."""
        return self.forwardDeclarationParser.lookup(self.original)

    def __eq__(self, other):
        if isinstance(other, AvroPlaceholder):
            return self.avroType == other.avroType
        elif isinstance(other, AvroType):
            return self.avroType == other
        else:
            return False

    def __hash__(self):
        return hash(self.avroType)

    def __repr__(self):
        """Represents the placeholder as its resolved type in JSON or ``{"type": "unknown"}`` if not resolved yet."""
        if self.forwardDeclarationParser.contains(self.original):
            return repr(self.forwardDeclarationParser.lookup(self.original))
        else:
            return '{"type": "unknown"}'

    def toJson(self):
        """Represent the resolved type as a JSON string."""
        return json.dumps(self.jsonNode())

    def jsonNode(self, memo=set()):
        """Represent the resolved type as a Jackson node.

        :type memo: set of strings
        :param memo: used to avoid infinite loops with recursive records
        """
        return self.avroType.jsonNode(memo)

    @property
    def parser(self):
        """The ``titus.datatype.ForwardDeclarationParser`` responsible for this placeholder."""
        return self.forwardDeclarationParser

class AvroFilledPlaceholder(AvroPlaceholder):
    """Used to create titus.datatype.AvroPlaceholder objects to satisfy functions that require them, yet the type is already known."""

    def __init__(self, avroType):
        self._avroType = avroType

    @property
    def avroType(self):
        return self._avroType

    def __repr__(self):
        return repr(self.avroType)

def parseAvroType(obj):
    """Parse an AVSC object without any memory of named types."""
    return schemaToAvroType(avro.schema.SchemaFromJSONData(obj, avro.schema.Names()))

class ForwardDeclarationParser(object):
    """Container that stores Avro types as they're collected from a PFA file, returning titus.datatype.AvroPlaceholder objects, and then resolves those types independent of the order in which they were read from the file."""

    def __init__(self):
        self.names = avro.schema.Names()
        self.lookupTable = {}

    def contains(self, original):
        return original in self.lookupTable

    def lookup(self, original):
        return self.lookupTable[original]

    @property
    def compiledTypes(self):
        return [x for x in self.lookupTable if isinstance(x, (AvroFixed, AvroRecord, AvroEnum))]

    def parse(self, jsonStrings):
        schemae = {}
        unresolvedSize = -1
        lastUnresolvedSize = -1
        errorMessages = {}

        while unresolvedSize != 0:
            for jsonString in jsonStrings:
                if jsonString not in schemae:
                    obj = json.loads(jsonString)

                    if isinstance(obj, str) and self.names.has_name(obj, None):
                        gotit = self.names.get_name(obj, None)
                        schemae[jsonString] = gotit
                    else:
                        oldnames = dict(self.names.names)

                        try:
                            gotit = avro.schema.SchemaFromJSONData(obj, self.names)
                        except avro.schema.SchemaParseException as err:
                            self.names = avro.schema.Names(names = oldnames)
                            errorMessages[jsonString] = str(err)
                        else:
                            schemae[jsonString] = gotit

            unresolved = [x for x in jsonStrings if x not in schemae]
            unresolvedSize = len(unresolved)

            if unresolvedSize == lastUnresolvedSize:
                raise titus.errors.SchemaParseException("Could not resolve the following types:\n    " +
                    "\n    ".join(["{0} ({1})".format(x, errorMessages[x]) for x in jsonStrings if x not in schemae]))
            else:
                lastUnresolvedSize = unresolvedSize

        result = dict((x, schemaToAvroType(schemae[x])) for x in jsonStrings)
        self.lookupTable.update(result)
        return result

    def getSchema(self, description):
        result = self.getAvroType(description)
        if result is None:
            return None
        else:
            return result.avroType

    def getAvroType(self, description):
        if isinstance(description, str):
            if self.names.has_name(description, None):
                return schemaToAvroType(self.names.get_name(description, None))
            elif description == "null":
                return AvroNull()
            elif description == "boolean":
                return AvroBoolean()
            elif description == "int":
                return AvroInt()
            elif description == "long":
                return AvroLong()
            elif description == "float":
                return AvroFloat()
            elif description == "double":
                return AvroDouble()
            elif description == "bytes":
                return AvroBytes()
            elif description == "string":
                return AvroString()
            else:
                try:
                    obj = json.loads(description)
                except ValueError:
                    return None
                else:
                    return self.getAvroType(obj)
        elif isinstance(description, dict):
            if description == {"type": "null"}:
                return AvroNull()
            elif description == {"type": "boolean"}:
                return AvroBoolean()
            elif description == {"type": "int"}:
                return AvroInt()
            elif description == {"type": "long"}:
                return AvroLong()
            elif description == {"type": "float"}:
                return AvroFloat()
            elif description == {"type": "double"}:
                return AvroDouble()
            elif description == {"type": "bytes"}:
                return AvroBytes()
            elif description == {"type": "string"}:
                return AvroString()
            elif description.get("type") == "array" and "items" in description:
                return AvroArray(self.getAvroType(description["items"]))
            elif description.get("type") == "map" and "values" in description:
                return AvroArray(self.getAvroType(description["values"]))
            elif description.get("type") in ("fixed", "enum", "record"):
                if self.names.has_name(description.get("name"), description.get("namespace")):
                    return schemaToAvroType(self.names.get_name(description.get("name"), description.get("namespace")))
                else:
                    raise titus.errors.AvroException("new types, like {0}, cannot be defined with the parser.getAvroType or parser.getSchema methods (use parse)".format(json.dumps(description)))
        elif isinstance(description, (tuple, list)):
            return AvroUnion([self.getAvroType(x) for x in description])
        else:
            return None

class AvroTypeBuilder(object):
    """Factory that coordinates the process of collecting Avro strings, putting them in a titus.datatype.ForwardDeclarationParser, and then resolving them all at the end, independent of order."""

    def __init__(self):
        self.forwardDeclarationParser = ForwardDeclarationParser()
        self.originals = []

    def removeEmptyNamespaces(self, x):
        if isinstance(x, dict):
            if "namespace" in x and (x["namespace"] is None or x["namespace"] == ""):
                del x["namespace"]
            return dict((k, self.removeEmptyNamespaces(v)) for k, v in list(x.items()))

        elif isinstance(x, (list, tuple)):
            return [self.removeEmptyNamespaces(v) for v in x]

        else:
            return x

    def removeDuplicateNames(self, x, memo):
        if isinstance(x, dict) and "name" in x and "type" in x and x["type"] in ("enum", "fixed", "record"):
            if "namespace" in x:
                name = x["namespace"] + "." + x["name"]
            else:
                name = x["name"]
            if name in memo:
                if memo[name] != x:
                    raise titus.errors.AvroException("type name \"{0}\" previously defined as\n{1}\nnow defined as\n{2}".format(name, ts(jsonNodeToAvroType(memo[name])), ts(jsonNodeToAvroType(x))))
                return name
            else:
                memo[name] = x
                return dict((k, self.removeDuplicateNames(v, memo)) for k, v in list(x.items()))

        elif isinstance(x, dict):
            return dict((k, self.removeDuplicateNames(v, memo)) for k, v in list(x.items()))

        elif isinstance(x, (list, tuple)):
            return [self.removeDuplicateNames(v, memo) for v in x]

        else:
            return x

    def makePlaceholder(self, avroJsonString, memo=None):
        if memo is not None:
            avroJsonString = json.dumps(self.removeDuplicateNames(self.removeEmptyNamespaces(json.loads(avroJsonString)), memo))
        self.originals.append(avroJsonString)
        return AvroPlaceholder(avroJsonString, self.forwardDeclarationParser)

    def resolveTypes(self):
        self.forwardDeclarationParser.parse(self.originals)
        self.originals = []

    def resolveOneType(self, avroJsonString):
        return ForwardDeclarationParser().parse([avroJsonString])[avroJsonString]

########################### Avro-Python is missing a JSON decoder, encoder, and comparator

def jsonDecoder(avroType, value):
    """Decode a JSON object as a given titus.datatype.AvroType.

    :type avroType: titus.datatype.AvroType
    :param avroType: how we want to interpret this JSON
    :type value: dicts, lists, strings, numbers, ``True``, ``False``, ``None``
    :param value: the JSON object in Python encoding
    :rtype: dicts, lists, strings, numbers, ``True``, ``False``, ``None``
    :return: an object ready for PFAEngine.action
    """

    if isinstance(avroType, AvroNull):
        if value is None:
            return value
    elif isinstance(avroType, AvroBoolean):
        if value is True or value is False:
            return value
    elif isinstance(avroType, AvroInt):
        try:
            return int(value)
        except (ValueError, TypeError):
            pass
    elif isinstance(avroType, AvroLong):
        try:
            return int(value)
        except (ValueError, TypeError):
            pass
    elif isinstance(avroType, AvroFloat):
        try:
            return float(value)
        except (ValueError, TypeError):
            pass
    elif isinstance(avroType, AvroDouble):
        try:
            return float(value)
        except (ValueError, TypeError):
            pass
    elif isinstance(avroType, AvroBytes):
        if isinstance(value, str):
            return value
    elif isinstance(avroType, AvroFixed):
        if isinstance(value, str):
            if len(value) == avroType.size:
                return value
    elif isinstance(avroType, AvroString):
        if isinstance(value, str):
            return value
    elif isinstance(avroType, AvroEnum):
        if isinstance(value, str) and value in avroType.symbols:
            return value
    elif isinstance(avroType, AvroArray):
        if isinstance(value, (list, tuple)):
            return [jsonDecoder(avroType.items, x) for x in value]
    elif isinstance(avroType, AvroMap):
        if isinstance(value, dict):
            return dict((k, jsonDecoder(avroType.values, v)) for k, v in list(value.items()))
    elif isinstance(avroType, AvroRecord):
        if isinstance(value, dict):
            out = {}
            for field in avroType.fields:
                if field.name in value:
                    out[field.name] = jsonDecoder(field.avroType, value[field.name])
                elif field.default is not None:
                    out[field.name] = jsonDecoder(field.avroType, field.default)
                elif isinstance(field.avroType, AvroNull):
                    out[field.name] = None
                else:
                    raise titus.errors.AvroException("{0} does not match schema {1}".format(json.dumps(value), ts(avroType)))
            return out
    elif isinstance(avroType, AvroUnion):
        if isinstance(value, dict) and len(value) == 1:
            tag, = list(value.keys())
            val, = list(value.values())
            types = dict((x.name, x) for x in avroType.types)
            if tag in types:
                return {tag: jsonDecoder(types[tag], val)}
        elif value is None and "null" in [x.name for x in avroType.types]:
            return None
    else:
        raise Exception
    raise titus.errors.AvroException("{0} does not match schema {1}".format(json.dumps(value), ts(avroType)))

def jsonEncoder(avroType, value, tagged=True):
    """Encode an object as JSON, given titus.datatype.AvroType.

    :type avroType: titus.datatype.AvroType
    :param avroType: type of this object
    :type value: dicts, lists, strings, numbers, ``True``, ``False``, ``None``
    :param value: the object returned from PFAEngine.action
    :type tagged: bool
    :param tagged: if True, represent unions as ``{tag: value}``; if False, represent them simply as ``value``.
    :rtype: dicts, lists, strings, numbers, ``True``, ``False``, ``None``
    :return: the JSON object in Python encoding
    """

    if isinstance(avroType, AvroNull) and value is None:
        return value
    elif isinstance(avroType, AvroBoolean) and (value is True or value is False):
        return value
    elif isinstance(avroType, AvroInt) and isinstance(value, int) and value is not True and value is not False:
        return value
    elif isinstance(avroType, AvroLong) and isinstance(value, int) and value is not True and value is not False:
        return value
    elif isinstance(avroType, AvroFloat) and isinstance(value, (int, float)) and value is not True and value is not False:
        return float(value)
    elif isinstance(avroType, AvroDouble) and isinstance(value, (int, float)) and value is not True and value is not False:
        return float(value)
    elif isinstance(avroType, AvroBytes) and isinstance(value, str):
        return value
    elif isinstance(avroType, AvroFixed) and isinstance(value, str):
        if len(value) == avroType.size:
            return value
    elif isinstance(avroType, AvroString) and isinstance(value, str):
        return value
    elif isinstance(avroType, AvroEnum) and isinstance(value, str) and value in avroType.symbols:
        return value
    elif isinstance(avroType, AvroArray) and isinstance(value, (list, tuple)):
        return [jsonEncoder(avroType.items, x, tagged) for x in value]
    elif isinstance(avroType, AvroMap) and isinstance(value, dict):
        return dict((k, jsonEncoder(avroType.values, v, tagged)) for k, v in list(value.items()))
    elif isinstance(avroType, AvroRecord) and isinstance(value, dict):
        out = {}
        for field in avroType.fields:
            if field.name in value:
                out[field.name] = jsonEncoder(field.avroType, value[field.name], tagged)
            elif field.default is not None:
                pass
            else:
                raise titus.errors.AvroException("{0} does not match schema {1}".format(json.dumps(value), ts(avroType)))
        return out
    elif isinstance(avroType, AvroUnion) and any(isinstance(t, AvroNull) for t in avroType.types) and value is None:
        return None
    elif isinstance(avroType, AvroUnion):
        if isinstance(value, dict) and len(value) == 1:
            val, = list(value.values())
            for t in avroType.types:
                try:
                    out = jsonEncoder(t, val, tagged)
                except titus.errors.AvroException:
                    pass
                else:
                    if tagged:
                        return {t.name: out}
                    else:
                        return out
        for t in avroType.types:
            try:
                out = jsonEncoder(t, value, tagged)
            except titus.errors.AvroException:
                pass
            else:
                if tagged:
                    return {t.name: out}
                else:
                    return out
    raise titus.errors.AvroException("{0} does not match schema {1}".format(json.dumps(value), ts(avroType)))

def compare(avroType, x, y):
    """Returns -1, 0, or 1 depending on whether x is less than, equal to, or greater than y, according to the schema.

    Assumes that x and y are valid examples of the schema.

    :type avroType: titus.datatype.AvroType
    :param avroType: type of this object
    :type x: dicts, lists, strings, numbers, ``True``, ``False``, ``None``
    :param x: Avro object in Python form
    :type y: dicts, lists, strings, numbers, ``True``, ``False``, ``None``
    :param y: Avro object in Python form
    :rtype: int
    :return: -1, 0, or 1
    """

    if isinstance(avroType, AvroNull) and x is None and y is None:
        return 0
    elif isinstance(avroType, AvroBoolean) and (x is True or x is False) and (y is True or y is False):
        return (x > y) - (x < y)   # agrees with Java
    elif isinstance(avroType, AvroInt) and isinstance(x, int) and x is not True and x is not False and isinstance(y, int) and y is not True and y is not False:
        return (x > y) - (x < y)
    elif isinstance(avroType, AvroLong) and isinstance(x, int) and x is not True and x is not False and isinstance(y, int) and y is not True and y is not False:
        return (x > y) - (x < y)
    elif isinstance(avroType, AvroFloat) and isinstance(x, (int, float)) and x is not True and x is not False and isinstance(y, (int, float)) and y is not True and y is not False:
        return (x > y) - (x < y)
        if math.isnan(x):
            if math.isnan(y):
                return 0
            else:
                return 1
        else:
            if math.isnan(y):
                return -1
            else:
                return (x > y) - (x < y)
    elif isinstance(avroType, AvroDouble) and isinstance(x, (int, float)) and x is not True and x is not False and isinstance(y, (int, float)) and y is not True and y is not False:
        if math.isnan(x):
            if math.isnan(y):
                return 0
            else:
                return 1
        else:
            if math.isnan(y):
                return -1
            else:
                return (x > y) - (x < y)
    elif isinstance(avroType, AvroBytes) and isinstance(x, str) and isinstance(y, str):
        return (x > y) - (x < y)
    elif isinstance(avroType, AvroFixed) and isinstance(x, str) and isinstance(y, str):
        return (x > y) - (x < y)
    elif isinstance(avroType, AvroString) and isinstance(x, str) and isinstance(y, str):
        return (x > y) - (x < y)
    elif isinstance(avroType, AvroEnum) and isinstance(x, str) and x in avroType.symbols and isinstance(y, str) and y in avroType.symbols:
        comparison = avroType.symbols.index(x) - avroType.symbols.index(y)
        if comparison < 0:
            return -1
        elif comparison > 0:
            return 1
        else:
            return 0
    elif isinstance(avroType, AvroArray) and isinstance(x, (list, tuple)) and isinstance(y, (list, tuple)):
        for xi, yi in zip(x, y):
            comparison = compare(avroType.items, xi, yi)
            if comparison != 0:
                return comparison
        if len(x) > len(y):
            return 1
        elif len(x) < len(y):
            return -1
        else:
            return 0
    elif isinstance(avroType, AvroMap) and isinstance(x, dict) and isinstance(y, dict):
        raise NotImplementedError("Avro has no order defined for maps???")
    elif isinstance(avroType, AvroRecord) and isinstance(x, dict) and isinstance(y, dict):
        for field in avroType.fields:
            if field.order == "ignore":
                continue
            comparison = compare(field.avroType, x[field.name], y[field.name])
            if comparison != 0:
                if field.order == "descending":
                    return -comparison
                else:
                    return comparison
        return 0
    elif isinstance(avroType, AvroUnion):
        if isinstance(x, dict) and len(x) == 1:
            (xtag, x), = list(x.items())
            xtypei, xtype = [(ti, t) for ti, t in enumerate(avroType.types) if t.name == xtag][0]
        else:
            xtypei = None
            for ti, t in enumerate(avroType.types):
                try:
                    jsonEncoder(t, x)
                except titus.errors.AvroException:
                    pass
                else:
                    xtypei = ti
                    xtype = t
            if xtypei is None:
                raise titus.errors.AvroException()
        if isinstance(y, dict) and len(y) == 1:
            (ytag, y), = list(y.items())
            ytypei, ytype = [(ti, t) for ti, t in enumerate(avroType.types) if t.name == ytag][0]
        else:
            ytypei = None
            for ti, t in enumerate(avroType.types):
                try:
                    jsonEncoder(t, y)
                except titus.errors.AvroException:
                    pass
                else:
                    ytypei = ti
                    ytype = t
            if ytypei is None:
                raise titus.errors.AvroException()
        if xtypei == ytypei:
            return compare(xtype, x, y)
        else:
            comparison = xtypei - ytypei
            if comparison < 0:
                return -1
            elif comparison > 0:
                return 1
            else:
                return 0
    else:
        raise titus.errors.AvroException("{0} or {1} does not match schema {2}".format(json.dumps(x), json.dumps(y), ts(avroType)))

########################### check data value against type

try:
    import numpy
    booleanTypes = (numpy.bool_,)
    integerTypes = (numpy.int_,
                    numpy.intc,
                    numpy.intp,
                    numpy.int8,
                    numpy.int16,
                    numpy.int32,
                    numpy.int64,
                    numpy.uint8,
                    numpy.uint16,
                    numpy.uint32,
                    numpy.uint64)
    floatTypes   = integerTypes + \
                   (numpy.float_,
                    numpy.float16,
                    numpy.float32)
except ImportError:
    booleanTypes = ()
    integerTypes = ()
    floatTypes   = ()

def checkData(data, avroType):
    """Return ``True`` if ``data`` satisfies ``avroType`` and can be used in PFAEngine.action."""

    if isinstance(avroType, AvroNull):
        if data == "null":
            data = None
        elif data is None:
            return data
        else:
            raise TypeError("expecting {0}, found {1}".format(ts(avroType), data))

    elif isinstance(avroType, AvroBoolean):
        if data == "true":
            return True
        elif data == "false":
            return False
        elif isinstance(data, booleanTypes):
            return bool(data)
        elif data is True or data is False:
            return data
        else:
            raise TypeError("expecting {0}, found {1}".format(ts(avroType), data))

    elif isinstance(avroType, AvroInt):
        if isinstance(data, str):
            try:
                data = int(data)
            except ValueError:
                raise TypeError("expecting {0}, found {1}".format(ts(avroType), data))
        elif isinstance(data, integerTypes):
            data = int(data)
        elif isinstance(data, int):
            return data
        else:
            raise TypeError("expecting {0}, found {1}".format(ts(avroType), data))

    elif isinstance(avroType, AvroLong):
        if isinstance(data, str):
            try:
                data = int(data)
            except ValueError:
                raise TypeError("expecting {0}, found {1}".format(ts(avroType), data))
        elif isinstance(data, integerTypes):
            data = int(data)
        elif isinstance(data, int):
            return data
        else:
            raise TypeError("expecting {0}, found {1}".format(ts(avroType), data))

    elif isinstance(avroType, AvroFloat):
        if isinstance(data, str):
            try:
                data = float(data)
            except ValueError:
                raise TypeError("expecting {0}, found {1}".format(ts(avroType), data))
        elif isinstance(data, floatTypes):
            data = float(data)
        elif isinstance(data, int):
            data = float(data)
        elif isinstance(data, float):
            return data
        else:
            raise TypeError("expecting {0}, found {1}".format(ts(avroType), data))

    elif isinstance(avroType, AvroDouble):
        if isinstance(data, str):
            try:
                data = float(data)
            except ValueError:
                raise TypeError("expecting {0}, found {1}".format(ts(avroType), data))
        elif isinstance(data, floatTypes):
            return float(data)
        elif isinstance(data, int):
            return float(data)
        elif isinstance(data, float):
            return data
        else:
            raise TypeError("expecting {0}, found {1}".format(ts(avroType), data))

    elif isinstance(avroType, (AvroBytes, AvroFixed)):
        if isinstance(data, bytes):
            return data
        elif isinstance(data, str):
            return stringToBytes(data)
        else:
            raise TypeError("expecting {0}, found {1}".format(ts(avroType), data))

    elif isinstance(avroType, (AvroString, AvroEnum)):
        if isinstance(data, str):
            return data
        else:
            raise TypeError("expecting {0}, found {1}".format(ts(avroType), data))

    elif isinstance(avroType, AvroArray):
        if hasattr(data, "__iter__"):
            return [checkData(x, avroType.items) for x in data]
        else:
            raise TypeError("expecting {0}, found {1}".format(ts(avroType), data))

    elif isinstance(avroType, AvroMap):
        if hasattr(data, "__iter__") and hasattr(data, "__getitem__"):
            newData = {}
            for key in data:
                value = checkData(data[key], avroType.values)
                if isinstance(key, bytes):
                    newData[key.decode("utf-8", "replace")] = value
                elif isinstance(key, str):
                    newData[key] = value
                else:
                    raise TypeError("expecting {0}, found key {1}".format(ts(avroType), key))
            return newData
        else:
            raise TypeError("expecting {0}, found {1}".format(ts(avroType), data))

    elif isinstance(avroType, AvroRecord):
        if hasattr(data, "__iter__") and hasattr(data, "__getitem__"):
            newData = {}
            for field in avroType.fields:
                try:
                    value = data[field.name]
                except KeyError:
                    raise TypeError("expecting {0}, couldn't find key {1}".format(ts(avroType), field.name))
                newData[field.name] = checkData(value, field.avroType)
            return newData
        else:
            raise TypeError("expecting {0}, found {1}".format(ts(avroType), data))

    elif isinstance(avroType, AvroUnion):
        if isinstance(data, dict) and len(data) == 1:
            tag, = list(data.keys())
            value, = list(data.values())
            for tpe in avroType.types:
                if tpe.name == tag:
                    if tag == "null":
                        return checkData(value, tpe)
                    else:
                        return {tag: checkData(value, tpe)}
            raise TypeError("expecting {0}, found {1}".format(ts(avroType), data))

        for tpe in avroType.types:
            try:
                newData = checkData(data, tpe)
            except TypeError:
                pass
            else:
                if tpe.name == "null":
                    return newData
                else:
                    return {tpe.name: newData}
        raise TypeError("expecting {0}, found {1}".format(ts(avroType), data))

    return data



#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

from titus.fcn import Fcn
from titus.fcn import LibFcn
from titus.signature import Sig
from titus.signature import Sigs
from titus.signature import Lifespan
from titus.signature import PFAVersion
from titus.datatype import *
from titus.errors import *
from titus.util import callfcn, div
import titus.P as P

provides = {}
def provide(fcn):
    provides[fcn.name] = fcn

prefix = "model.svm."

#################################################################### 
class Score(LibFcn):
    name = prefix + "score"
    sig = Sig([{"datum": P.Array(P.Double())},
               {"model": P.WildRecord("L",
                 {"const": P.Double(), 
                  "posClass": P.Array(P.WildRecord("M", {"supVec": P.Array(P.Double()), "coeff" : P.Double()})),
                  "negClass": P.Array(P.WildRecord("N", {"supVec": P.Array(P.Double()), "coeff" : P.Double()}))})},
	       {"kernel": P.Fcn([P.Array(P.Double()), P.Array(P.Double())], P.Double())}
               ], P.Double())
    errcodeBase = 12000
    def __call__(self, state, scope, pos, paramTypes, datum, model, kernel):
        const    = model["const"]
        negClass = model["negClass"]
        posClass = model["posClass"]
        if len(negClass) == 0 and len(posClass) == 0:
            raise PFARuntimeException("no support vectors", self.errcodeBase + 0, self.name, pos)
        negClassScore = 0.0
        for sv in negClass:
            supVec = sv["supVec"]
            if len(supVec) != len(datum):
                raise PFARuntimeException("support vectors must have same length as datum", self.errcodeBase + 1, self.name, pos)
            coeff  = sv["coeff"]
            negClassScore += callfcn(state, scope, kernel, [supVec, datum])*coeff
        posClassScore = 0.0
        for sv in posClass:
            supVec = sv["supVec"]
            if len(supVec) != len(datum):
                raise PFARuntimeException("support vectors must have same length as datum", self.errcodeBase + 1, self.name, pos)
            coeff  = sv["coeff"]
            posClassScore += callfcn(state, scope, kernel, [supVec, datum])*coeff
        return negClassScore + posClassScore + const
provide(Score())
